//****************************************************************************
//
//
//        The Minus4 project - Copyright (c) Mike Dailly 1996,97,98
//
//
// This source and the Minus4 project remains the copyright of Mike Dailly, 
// but can be freely used and changed for NON-Commercial purposes. 
//
//                       Keep emulators free!
//
// Remember if you add or change anything, you MUST send a copy
// complete with source to me at mike@scotch.demon.co.uk
// If its just a bug fix, then just send the changed file, with a comment
// telling me WHERE you fixed and WHAT you fixed.
//
// 
//
//                              6502.C
//                                      
//              This file contains the 6502 emulation
//
// This source
//
//
// Version  By   Date     Description
//
// v1.00    MD 08/04/96   All Instructions emulated in high level
// V1.2     MD 28/07/97   Perhaps I should have kept up this list...
//                        I'll try in future.                
// V1.21    MD 26/09/97   Added "OAL" and another "SKB" instruction
//                                      
//
// Last screen raster = $137 (311) - checked on the Plus 4
// Cycles per scanline= (slow)57  (fast) 114 - 5 RAM refresh = 109
//
//****************************************************************************
#include "headers\md.h"
#include "headers\all.h"
#include "headers\screen.h"
#include "headers\draw.h"
#include "headers\6502asm.h"
#include "headers\disasm.h"
#include "headers\keyb.h"
#include "headers\keyboard.h"
#include "headers\sound.h"
#include "headers\sb.h"
#include "headers\disk.h"
#include "headers\wind.h"
#include "headers\main.h"
#include "headers\soundb.h"



#define  framecount  SCREEN_LINES*CYCLES_PER_LINE

#define	BREAK_PC		1
#define	SKIP_PC		2
#define	BREAK_MEM	4


#define  word  unsigned short
#define  byte  unsigned char
#define  CHECKKEY \
           key=kbhit(); \
           if (key!=0x0) break;

// 8501
#define _C     1
#define _Z     2
#define _I     4
#define _D     8
#define _B     16
#define _V     64
#define _N     128


// *******************************************************************
// Tape loader constants
// *******************************************************************
#define  _1MHZ          0
#define  _1_78MHZ       1
#define	TONESIZE	144

#define	FPS							50.0f
#define	LINES							312.0f
#define	DOT_CLOCK					228.0f
#define	TAPE_SAMPLE_RATE			44100.0f

#define	TAPE_COUNT					((DOT_CLOCK*LINES*FPS)/TAPE_SAMPLE_RATE)




// *******************************************************************
//
// externals
//
// *******************************************************************
extern	uint8		Tone[TONESIZE];	// TED "tone" at middle "C"

extern	uint32	DOCOPYSCREEN;
extern   uint8 col;
extern   unsigned char     _A;
extern   unsigned char     _X;
extern   unsigned char     _Y;
extern   unsigned char     _F;
extern   unsigned char     _SP;
extern   uint32            _PC;  
extern   uint32   VBLANK;
extern   uint8    *pBuffer1;  // TED samples
extern   uint8    *pBuffer2;  // SID channel 1
extern   uint8    *pBuffer3;  // SID channel 2
extern   uint8    *pBuffer4;  // SID channel 3
extern   uint8    *pBuffer5;  // SID samples
extern   uint8    *pBuffer6;
extern   uint8    *pBuffer7;
extern   uint8    *pBuffer8;  // output "mixed" channel
extern   uint32   FIndex;
extern   tWindow  *pMaster;
extern	char 		C_Table[256];
extern	tWindow	*pLoading;


	uint32	ScreenBlank;
	uint8		TEDSample[256];
	uint32	ScaleScreen=0;
	uint32	MidSample=0;
	uint32	SoundShifter=0;

   uint32   CYCLES_PER_LINE_ON=CYCLES_PER_LINE_ON_def;            // ++ NORMAL ON  SCREEN CYCLES ++ 
   uint32   CYCLES_PER_LINE_OFF=CYCLES_PER_LINE_OFF_def;       // ++ NORMAL OFF SCREEN CYCLES ++ 
   uint32   IRQ_START_CYCLE=7;   //4;?!?!
// uint32   IRQ_START_CYCLE=1;   //8;


   uint8    *pROMS[16];
   uint8    *ROMS[16];
   uint32   DRAW_LINE=2;         //2
   //uint32   DRAWSTART=0x1f0000;         //2
   uint32   DRAWSTART=0x000000;         // "0" (start of 38 char window)

   char     hexbuff[16]={"0123456789ABCDEF"};
   uint8    *pMemory;               // base address - non aligned.
   uint32   *pMemoryReadTable;
   uint32   *pMemoryWriteTable;
   uint32   *pMemoryRead;        
   uint32   *pMemoryWrite;    
   uint32   *pMemoryRead2;    
   uint8    *Memory;                // main memory
   uint32   *MemoryReadTable;
   uint32   *MemoryWriteTable;
   uint32   *MemoryRead;            // pointer to variables to flag when written to
   uint32   *MemoryWrite;           // pointer to variables to flag when written to
   uint32   *MemoryRead2;           // pointer to variables to flag when written to
   uint32   DummyWriteVariable;     // the dummy variable

	uint32	y1,y2;

	// *******************************************************************
	//
	// These strings are used to calculate "ff1e" lookup tables, and processor
	// cycles per line.
	//
	// End of horizontal blank is $D0, so $ff1e starts there each line
	//
	// p = Processor works
	// * = Processor can write
	// - = Empty
	// S = TED fetches memory (0th line=character number, 7th=Char attribute)
	// F = TED fetches FONT memeory
	// R = TED refreshes RAM
	//
	// *******************************************************************
	//char		FastDMA[114]    ="pppp-*-*-*-SFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFpRpRpRpRpRppppppppppppp";
	//char		SlowDMA[114]    ="-p-p-*-*-*-SFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFpRpRpRpRpRp-p-p-p-p-p-p";
	char		FastDMA[114]    ="pppp-p-p-p-SFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFpRpRpRpRpRppppppppppppp";
	char		SlowDMA[114]    ="-p-p-p-p-p-SFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFSFpRpRpRpRpRp-p-p-p-p-p-p";
	char		FastNormal[114] ="pppp-p-p-p-pFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpRpRpRpRpRppppppppppppp";
	char		SlowNormal[114] ="-p-p-p-p-p-pFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpFpRpRpRpRpRp-p-p-p-p-p-p";
	char		FastBoarder[114]="ppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppppRpRpRpRpRppppppppppppp";
	char		SlowBoarder[114]="-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-p-pRpRpRpRpRp-p-p-p-p-p-p";

	uint8		tfastdma[114];
	int32		cfastdma;
	uint8		tslowdma[114];
	int32		cslowdma;
	uint8		tfastnormal[114]; 
	int32		cfastnormal;
	uint8		tslownormal[114]; 
	int32		cslownormal;
	uint8		tfastboarder[114];
	int32		cfastboarder;
	uint8		tslowboarder[114];
	int32		cslowboarder;
	uint8		*pCurrentFF1E;
   uint8   	ff1e_old;         // char counter..

	// *******************************************************************
   //
	// *******************************************************************

	int32		FF1e_DONELINE=0;
	uint32	ResetRasterLine;
   uint32   Screen2,VidBase2;
   uint32   DoneSample=0;
	uint32	Halfway=0;
   uint32   _01count=100;

   uint32   testjoy=0;
   uint32   Joystick_active=0,joy_max_x=0;
   uint32   joy_max_y=0,joy_min_x=0xffffffff,joy_min_y=0xffffffff;
   uint32   joy_cen_x=0,joy_cen_y=0;
   uint32   Joy1=0;              // actual direction bits, 1=up, 2=down, 4=left, 8=right, 16=button

   uint8    skipit=0;
   uint8    CharLine=0;          // current line in character BITMAP being drawn!
   int32		CharRow=0;           // current character ROW being processed.

   uint32   FlashCount;
   uint32   cTimer1=0;                    // Current Timer (32 bit for 16 bit timer+FRACTION)
   uint32   cTimer2=0;                    // Current Timer
   uint32   cTimer3=0;                    // Current Timer
   uint32   Timer1=0;                     // timer init value
   uint32   Timer2=0;                     // timer init value
   uint32   Timer3=0;                     // timer init value
   uint32   TIMER1_LO=0;                    // value written to LOW?
   uint32   TIMER2_LO=0;                    // value written to LOW?
   uint32   TIMER3_LO=0;                    // value written to LOW?
   uint8    *pTimer1Lo=(&cTimer1)+2;
   uint8    *pTimer1Hi=(&cTimer1)+3;

   uint32   AND_IRQ=0;
   uint32   JOYSTICK;
   uint32   ROM;                 // rom active flag!
	uint8		RomIndex=0;				// index to activate ROM
   uint32   ROMINF;              // rom low number
   uint32   KEYBOARD_LATCH=0;    // accessed (written) keyboard latch!
   uint32   KEYBOARD_LATCH_2=0;  // accessed (written) keyboard latch!

	uint32	filter=1;
	
   uint32   BrkPoint[10];

   uint32   CPU_MODE=0;          // 0 = 1Mhz mode, Not 0 = 1.78Mhz
   uint8    TimerLow;
   uint8    TimerHigh;
   uint16      newpc;
   uint32      LAST_PC;
   uint8    ThisLine=0;
   int      _ret;
   int32    aa;
   uint32   charcount;
   uint32   LastLine;
   char     *pFileName;
   uint8    CharAttribute[40];            // read on the LAST line of the previous character
   uint8    CharNumber[40];               // read on the FIRST line of the character
   uint32   ScreenIndexArray[200*40];        // Screen index array...
   int32    mem_break;  
   int32    brk_address;   
   int32    do_break;   
   int32    _screen;
   uint32   NextCycles;
   uint32   NextCycle_base;
   uint32   Cycles;
   uint32   CycleCount;
   uint32   LineCycles=0;        // okay? :-)
   int32    ScreenLine;
   uint32   RasterLine;
   //uint32   opcode;
   int32		ToneCounter=0;

   uint32   r1,r2;
   uint32   MemDump;
   uint32   DoLine,DRAWIT;


   uint32   EXIT_PRG=0;
   uint32   CYCLES_PER_LINE=CYCLES_PER_LINE_ON_def;
   int32    scroffset=4;
   int32    IRQrast;
   int32   mem_ff1e;         // char counter..
   int32   mem_ff1e_old;         // char counter..
   int32   mem_ff1e_delta;

   uint32   DontInc;

   uint32   CycleIndex[255];
   byte  	data;
   uint16   _d1;

   uint32   TESTPEEK=0;
   uint8    f7;
	uint32	StartDisPC=0x10000,EndDisPC=0x10000;
	uint32	EditMem=0x10000,EditNibble=0;
	uint8		BreakPoints[65536];	// 
	uint32	SampleStart=0;
	int32		TapeCountCheck;
	int32		TapeSoFar;
	int32		TapeFileSize;

	int32		DrawNextRaster;

	
	// *******************************************************************
	//
	// Tape loading edge detection "screw"
	//
	// *******************************************************************
	uint8		TapeHighEdgeTable[5] ={0xf4,0xf8,0xe0,0xc0,0x80};
	uint8		TapeLowEdgeTable[5]  ={0x03,0x07,0x1f,0x3f,0x7f};
	uint8		TapeHighEdge=0xc0;			
	uint8		TapeLowEdge=0x03f;				
	int32		FASTLOADER=0;
	char		TapeFileName[1024];
	int32		TapeEdge=3;


uint8 GETABYTE( void );
#pragma aux GETABYTE =              \
      "mov  ecx,_PC"                \
      "mov  al,cl"                  \
      "shl  ecx,2"                  \
      "and  al,3"                   \
      "add  ecx,MemoryReadTable"    \
      "mov  ecx,[ecx]"              \
      "or   cl,al"                  \
      "inc  _PC"                    \
      "mov  cl,[ecx]"               \
      modify [ecx al]               \
      value [cl]
		uint8	*pHeader;
		uint8	*pHeader2;

uint16 GETAWORD( void );
#pragma aux GETAWORD =              \
      "mov  ecx,_PC"                \
      "mov  al,cl"                  \
      "shl  ecx,2"                  \
      "and  al,3"                   \
      "add  ecx,MemoryReadTable"    \
      "mov  ecx,[ecx]"              \
      "or   cl,al"                  \
      "add  word ptr _PC,2"         \
      "mov  cx,[ecx]"               \
      modify [ecx al]               \
      value [cx]


void DOCOMP( uint8 val );
#pragma aux DOCOMP =                \
   "and  eax,255"                   \
   "mov  ebx,dword ptr _F"          \
   "and  ebx,4+8+64+1+16"           \
   "or   bl,CompTable[eax]"         \
   "mov  dword ptr _F,ebx"          \
   parm caller [al]                 \
   modify [eax]                     


	// **********************************************
	// Window system
	// **********************************************
	uint16	bShadows[16]={0,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7};

typedef	struct	tTextWindow{
		int32		flags; 
		int32		x,y,xs,ys;
		uint32	ScreenAddress;
		uint8		*pMemory;
}tTextWindow;

	tTextWindow	Windows[20];
	int32			NextWindow;


	int8			Shifted=0;

	// **********************************************
	//
	// Keyboard for INPUT() routine
	//
	// **********************************************
	uint8			KeyVal[80] ={	2,3,4,5,6,7,8,9,10,11,
										16,17,18,19,20,21,22,23,24,25,
										30,31,32,33,34,35,36,37,38,
                              44,45,46,47,48,49,50,
										K_SEMICOLON,K_BACKSLASH,K_FULLSTOP,K_MINUS

									 };
	// NON shifted keyboard
	char			KeyChar[80]={	'1','2','3','4','5','6','7','8','9','0',
										'q','w','e','r','t','y','u','i','o','p',
										'a','s','d','f','g','h','j','k','l',
										'z','x','c','v','b','n','m',
										';','\\','.','-'
                            };
	// Shifted keyboard
	char			SKeyChar[80]={	'!','2','3','$','5','6','7','*','9','0',
										'q','w','e','r','t','y','u','i','o','p',
										'a','s','d','f','g','h','j','k','l',
										'z','x','c','v','b','n','m',
										':','|','>','_'
       							 };


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

int32	FindWindow( void )
{
	int32	i;

	for (i=0;i<20;i++){
		if ((Windows[i].flags&1)==0){
			return i;
		}
	}
	return 0xffffffff;
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

int32	OpenTextWindow(int32 x,int32 y, int32 xsize,int32 ysize, char *pName)
{
	int32		i,i2,Address,index;
	uint16	*pAddress,*pAdd,AttrChar;
   uint16	*pMem;

	NextWindow=FindWindow();

	Windows[NextWindow].flags|=1;
	Windows[NextWindow].pMemory=MyMalloc((xsize+1)*(ysize+1)*2);
	Windows[NextWindow].x=x;
	Windows[NextWindow].y=y;
	Windows[NextWindow].xs=xsize;
	Windows[NextWindow].ys=ysize;
	Windows[NextWindow].ScreenAddress=(0xb8000+(x*2)+(y*160));
	//
	// Save TEXT screen (under window)
	//
	pMem=(uint16*) Windows[NextWindow].pMemory;
	pAddress=(uint16*) Windows[NextWindow].ScreenAddress;
	for (i2=y;i2<y+ysize+1;i2++){
		pAdd=pAddress;
		for (i=x;i<x+xsize+1;i++){
			*pMem++=*pAdd++;
		}
		pAddress+=80;							// next line
	}

	// Print Boarder - TOP row first
	pAddress=(uint16*)Windows[NextWindow].ScreenAddress;
	pAdd=pAddress;
	*pAdd++=0x1fc9;
	for (i=x+1;i<x+xsize-1;i++){
		*pAdd++=0x1fcd;
	}
	*pAdd++=0x1fbb;

	// Main body of window
	pAdd=pAddress+=80;
	for (i2=y+1;i2<y+ysize-1;i2++){
		pAdd=pAddress;
	   *pAdd++=0x1fba;
	   for (i=x+1;i<x+xsize-1;i++){
		   *pAdd++=0x1f20;
	   }
	   *pAdd++=0x1fba;
	   *pAdd=(bShadows[(*pAdd&0xf000)>>12]<<12)|
            (bShadows[(*pAdd&0xf00)>>8]<<8)|
			   (*pAdd&0xff);
		pAddress+=80;
	}

	// And finaly the lower boarder
	pAdd=pAddress;
	*pAdd++=0x1fc8;
	for (i=x+1;i<x+xsize-1;i++){
		*pAdd++=0x1fcd;
	}
	*pAdd++=0x1fbc;
	*pAdd=(bShadows[(*pAdd&0xf000)>>12]<<12)|
         (bShadows[(*pAdd&0xf00)>>8]<<8)|
			(*pAdd&0xff);
	

	// Print lower shadow
	pAdd=pAddress+81;
	for (i=x+1;i<x+xsize+1;i++){
	   *pAdd++=	(bShadows[(*pAdd&0xf000)>>12]<<12)|
            	(bShadows[(*pAdd&0xf00)>>8]<<8)|
			   	(*pAdd&0xff);
 	}
	

	//
	// Now print the window name over the top (if it HAS one)
	//
	if (pName!=NULL){
	   pAddress=(uint16*)Windows[NextWindow].ScreenAddress;
		index=(xsize-Len(pName))/2;
		pAddress+=index;
	   for (i=Len(pName);i>0;i--){
			AttrChar=(*pName++)|0x1e00;
		   *pAddress++=AttrChar;
	   }
	}	

	return NextWindow;
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void CloseTextWindow (int32 handle)
{
	int32		i,i2,x,y,xsize,ysize;
	uint16	*pAddress,*pAdd,AttrChar,*pMem;


	Windows[NextWindow].flags=0;
	x=Windows[handle].x;
	y=Windows[handle].y;
	xsize=Windows[handle].xs;
	ysize=Windows[handle].ys;

	//
	// Restore TEXT screen
	//
	pMem=(uint16*) Windows[handle].pMemory;
	pAddress=(uint16*) Windows[handle].ScreenAddress;
	for (i2=y;i2<y+ysize+1;i2++){
		pAdd=pAddress;
		for (i=x;i<x+xsize+1;i++){
			*pAdd++=*pMem++;
		}
		pAddress+=80;							// next line
	}

	MyFree(Windows[handle].pMemory);
}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
char	FindKey(int32 v)
{
	int32	i;

	if (Shifted==0){
	   for (i=0;i<80;i++){
		   if	(KeyVal[i]==v){
          	return KeyChar[i];
			}
	   }
	}else{
	   for (i=0;i<80;i++){
		   if	(KeyVal[i]==v){
          	return SKeyChar[i];
			}
	   }
	}
	return 0;
}

//************************************************************
//                                                           *
//   Name    :	Input                                         *
//   Function: Input a string in TEXT mode                   *
//                                                           *
//   In   :	x = xcoord                                       *
//          y = ycoord                                       *
//          xsize = size of input window                     *
//          pName = Name of window                           *
//          MaxSize= MAX number of chars to input (<win size)*
//   Out  : char* = input string                             *
//                                                           *
//************************************************************

char	*Input(int32 x,int32 y,int32 xsize, char *pName, int32 MaxSize)
{
	int32	wind,i,NextChar=0,Print,index;
	char	c;
   char	*pChar;
	char	space[1024];
	uint8	Attribute;

	wipe();
	Attribute=TextColour;
	SetPrintCol(0x1f);		// Blue background, Yellow Text


	pChar=MyMalloc(1024);
	pChar[0]='_';
	pChar[1]=0x0;
	
   for (i=0;i<xsize-2;i++) space[i]=0x20;
	space[xsize-2]=0x0;
	wind=OpenTextWindow (x,y,xsize,5,pName);

	Print=1;
	while ((Keys2[K_ESCAPE]==0) AND (Keys2[K_ENTER]==0)){
		if (Print==1){
			Print=0;
      	PrintLine (x+1,y+2,&space);
			index=((xsize-2)-Len(pChar))/2;
      	PrintLine (x+1+index,y+2,pChar);
		}

		// Shift?
		if ((Keys[K_LEFTSHIFT]!=0) OR (Keys[K_RIGHTSHIFT]!=0)){
			Shifted=1;
			Keys2[K_LEFTSHIFT]=0;
			Keys2[K_RIGHTSHIFT]=0;
		} else{
			Shifted=0;
			Keys2[K_LEFTSHIFT]=0;
			Keys2[K_RIGHTSHIFT]=0;
		}


		// key pressed?
		for (i=0;i<255;i++){
			if (Keys2[i]!=0){
       	   if (i==K_DELETE){
				   if (NextChar>0){
				      NextChar--;
				      pChar[NextChar]='_';
				      pChar[NextChar+1]=0x0;
						Print=1;
				   }
			   }else{
				   c=FindKey(i);
				   if (c!=0){
					   if (NextChar!=MaxSize){
					      pChar[NextChar]=c;
					      pChar[NextChar+1]='_';
					      pChar[NextChar+2]=0x0;
					      NextChar++;
							Print=1;
					   }
				   }
				}
				Keys2[i]=0;
			}
		}
	}


	pChar[NextChar]=0x0;
	CloseTextWindow(wind);
	SetPrintCol(Attribute);		// Blue background, Yellow Text
   wipe();  
	return pChar;
}

//************************************************************
//                                                           *
//   Name    : AddSound/RemoveSound                          *
//   Function: Allow sound to be turned on and off           *
//                                                           *
//************************************************************
void AddSound( void )
{
   // 
   // If sound enabled, enable write detects. Otherwise leave it coz
   // we don't care..
   //
   if (SOUND==1){
      MemoryWrite[0xff0e]=(uint32) &Chanel1_Lo|ROUTINE;
      MemoryWrite[0xff0f]=(uint32) &Chanel2_Lo|ROUTINE;  
      MemoryWrite[0xff10]=(uint32) &Chanel2_Hi|ROUTINE;  
      MemoryWrite[0xff11]=(uint32) &Chanel_Select|ROUTINE;  
      MemoryWrite[0xff12]=(uint32) &Chanel1_Hi|ROUTINE;
   }
}
void RemoveSound( void )
{
   MemoryWrite[0xff0e]=((uint32) &Memory[0xff0e])&0xfffffffc;
   MemoryWrite[0xff0f]=((uint32) &Memory[0xff0f])&0xfffffffc;
   MemoryWrite[0xff10]=((uint32) &Memory[0xff10])&0xfffffffc;
   MemoryWrite[0xff11]=((uint32) &Memory[0xff11])&0xfffffffc;
   MemoryWrite[0xff12]=((uint32) &Memory[0xff12])&0xfffffffc;
	EnableNoise=0;		// disable "noise"
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

uint8 ReadJoystickButtons  ( void )
{
   outp     (JOYPORT,0);      // clear the latch and request a sample
   return   ( (inp(JOYPORT)&0xf0) );   // return all buttons
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void CalabrateJoystick ( void )
{
   char  s[128];
   uint32   X,Y;
   uint16   axis[4];
   uint8    but;
   tWindow  *pCal;
   tGadget  *txt1,*txt2,*txt3,*txt4,*txt5,*txt6,*txt7,*txt8;   
   tGadget  *txt9,*txt10,*txt11,*txt12,*txt13,*txt14;

   wipe();  // wipe keys pressed

   pCal=OpenWindow (NULL, 80,130,200,70);
   SetWindowPen (pCal,255,245,244,243,240);     // set to a green colour scheme
   txt1=AddText(pCal,5,13, "Wiggle joystick in ALL directions then");
   txt2=AddText(pCal,5,20, "release and press fire or ESCAPE");
   txt3=AddText(pCal,5,30, "Current X");
   txt4=AddText(pCal,5,37, "Max X    ");
   txt5=AddText(pCal,5,44, "Min X    ");
   txt6=AddText(pCal,55,30, "=       ");
   txt7=AddText(pCal,55,37, "=       ");
   txt8=AddText(pCal,55,44, "=       ");
   txt9=AddText(pCal,105,30, "Current Y");
   txt10=AddText(pCal,105,37, "Max Y    ");
   txt11=AddText(pCal,105,44, "Min Y    ");
   txt12=AddText(pCal,155,30, "=       ");
   txt13=AddText(pCal,155,37, "=       ");
   txt14=AddText(pCal,155,44, "=       ");
                          

   joy_max_x=0;
   joy_max_y=0;
   joy_min_x=0xffffffff;
   joy_min_y=0xffffffff;
   joy_cen_x=0;
   joy_cen_y=0;

   while ( ((ReadJoystickButtons()&0xf0)==0xf0) AND (Keys[K_ESCAPE]==0)) {

      ProcessInterface (0x0,xx,yy,but);               // Do all mouse clicks etc. and pass key press
      DoMouse ();
      WaitVBlank();                          // slow down a bit...
      delay (2);                             // just in case VBLANK doesn't work (some video cards)

      joyread(&axis[0],&but,3);
      X=axis[0];
      Y=axis[1];

      if (X>joy_max_x) joy_max_x=X;
      if (X<joy_min_x) joy_min_x=X;

      sprintf(s,"=%d ",X);
      strcpy (txt6->pMessage,s);
      sprintf(s,"=%d ",joy_max_x);
      strcpy (txt7->pMessage,s);
      sprintf(s,"=%d ",joy_min_x);
      strcpy (txt8->pMessage,s);
      txt6->flags|=gREFRESH;
      txt7->flags|=gREFRESH;
      txt8->flags|=gREFRESH;

      if (Y>joy_max_y) joy_max_y=Y;
      if (Y<joy_min_y) joy_min_y=Y;

      sprintf(s,"=%d ",Y);
      strcpy (txt12->pMessage,s);
      sprintf(s,"=%d ",joy_max_y);
      strcpy (txt13->pMessage,s);
      sprintf(s,"=%d ",joy_min_y);
      strcpy (txt14->pMessage,s);
      txt12->flags|=gREFRESH;
      txt13->flags|=gREFRESH;
      txt14->flags|=gREFRESH;

   
      RenderInterface ();                          // print windows!
      CopyScreen (1);
   }
   
   // user has released joystick, so get center
   joyread(&axis[0],&but,3);
   joy_cen_x=axis[0];
   joy_cen_y=axis[1];
   CloseWindow (pCal);

   wipe();  
// ScreenMode=BitMapScreen;
// InitScreen ();
// _screen=1;
   CopyScreen (1);
   Joystick_active=1;         // Been through calabrate at least once.
}
//***********************************************************
//                                                          *
//   Name    : ProcessJoyStick                              *
//   Function: This processes an "analoge" joystick.        *
//             This works by takeing a left,right and center*
//             reading, and then taking half of each        *
//             distance, this gives something like this.    *
//                                                          *
//                ------------------------                *
//                                                        *
//                                                        *
//                                                        *
//                                                        *
//                     ---------                        *
//                                                      *
//                                                      *
//                     Dead Zone                        *
//                                                      *
//                                                      *
//                     ---------                        *
//                                                        *
//                                                        *
//                                                        *
//                                                        *
//                                                        *
//                ------------------------                *
//                                                          *
//             When the values are within the dead zone,    *
//             there is no movement. Otherwise we set bits  *
//             in that direction.                           *
//                                                          *
//   In   : none                                            *
//   Out  : Joy1 holds direction(s)                         *
//                                                          *
//***********************************************************

#define  JOY_UP      1
#define  JOY_DOWN    2
#define  JOY_LEFT    4
#define  JOY_RIGHT   8
#define  JOY_BUTTON  16


void ProcessJoyStick ( void )
{
   uint32      X,Y;
   uint32      xd,yd;
   uint32      Lcx,Rcx,Tcy,Bcy;
   uint16      axis[4];
   uint8       but;

   if (Joystick_active==0) return;     // no joystick attached.
   if ((JOYSTICK&3)==0) return;        // no joysticks active


   joyread(&axis[0],&but,3);
   X=axis[0];
   Y=axis[1];

   Lcx=(joy_cen_x-joy_min_x)/2;        // get distance from LEFT to CENTER and div 2
   Rcx=(joy_max_x-joy_cen_x)/2;
   Tcy=(joy_cen_y-joy_min_y)/2;        // get distance from TOP to CENTER and div 2
   Bcy=(joy_max_y-joy_cen_y)/2;


   Joy1=0;
   if (X< (joy_cen_x-Lcx) ) Joy1+=JOY_LEFT;
   if (X> (joy_cen_x+Rcx) ) Joy1+=JOY_RIGHT;

   if (Y< (joy_cen_y-Tcy) ) Joy1+=JOY_UP;
   if (Y> (joy_cen_y+Bcy) ) Joy1+=JOY_DOWN;

   if (ReadJoystickButtons()!=0xf0) Joy1+=JOY_BUTTON;

}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void  DoWindowStuff(char a)
{
   ProcessInterface (a,xx,yy,but);              // Do all mouse clicks etc. and pass key press

	if (pMaster!=NULL){
      // Only process gadgets if something in the windows was pressed!
      if ((pMaster->flags&wACTION)!=0){

         // ************************
         //
         // Calabrate joystick
         //
         // ************************
         if (txt3->value!=0){
            CalabrateJoystick ();
            txt3->value=0;                            // clear button pressed
            txt3->flags|=gREFRESH;                    // refresh gadget
         }

         // ************************
         //
         // Enable/Disable joysticks
         //
         // ************************
         if (pJoystick1->value==0){                // process joysticks
            JOYSTICK&=~1;
         } else {
            JOYSTICK|=1;
         }

         if (pJoystick2->value==0){
            JOYSTICK&=~2;
         } else {
            JOYSTICK|=2;
         }

         // ************************
         //
         // Switch SOUND on/off
         //
         // ************************
         if ((txt6->value==0) AND (SOUND==1)){
            SOUND=0;
            ShutDownSound();
            RemoveSound();

         } else if ((txt6->value!=0) AND (SOUND==0)){
            SOUND=1;
            InitSound();
            AddSound();
         }

         // ************************
         //
         // Switch SAMPLES on/off
         //
         // ************************
         if ((txt10->value==0) AND (SB_ON==1)){
            DisableSB ();
            SB_ON=0;
            InitSB ();
         } else if ((txt10->value!=0) AND (SB_ON==0)){
            DisableSB ();
            SB_ON=1;
				filter=1;
            InitSB ();
         }

         // ************************
         //
         // Switch TIMER on/off
         //
         // ************************
         if ((txt4->value==0) AND (TIMER_ON==1)){
            TIMER_ON=0;
            TimerOff();
         } else if ((txt4->value!=0) AND (TIMER_ON==0)){
            TIMER_ON=1;
            TimerON();
         }

         // ************************
         //
         // Switch VBlank syncing
         //
         // ************************
         if (txt8->value==0){
            DOVBLANK=0;
         } else {
            DOVBLANK=1;
         }

      }     

      if ((pMaster->flags&wCLOSE)!=0){
		   CloseWindow(pMaster);
	      pMaster=NULL;
         //pMaster->flags&=~wCLOSE;
         //pMaster->Options|=HIDE;       // disable window system
         EnableMouse=0;
      }
   }
	
	if ((FileLoading==1)){
		PrintLoading();
	}
   RenderInterface ();                          // print windows!
}




//************************************************************
//                                                           *
//   Name    : FreezeGame                                    *
//   Function: Save the contents of memory, and any hardware *
//             info we will need to restart                  *
//                                                           *
//   In   : none (name?)                                     *
//   Out  : none                                             *
//                                                           *
//************************************************************
void FreezeGame( void )
{
   int      handle;
   uint16   Add;
	uint16	version=1;		// set version

   handle=open ("test.fre",O_WRONLY|O_BINARY|O_CREAT,S_IRWXU|S_IRWXG);
   Add=0x0000;
   write (handle,&Add,2);                 	// write load address

	write (handle,&version,2);						// write FREEZE version

   write (handle,&_PC,4);                    // write register
   write (handle,&_F,4);                     // write register
   write (handle,&_A,4);                     // write register
   write (handle,&_X,4);                     // write register
   write (handle,&_Y,4);                     // write register
   write (handle,&_SP,4);                    // write register

   write (handle,&ROMINF,4);                 // write ROM bank data!
   write (handle,&ROM,4);                    // ROM on/off?
   write (handle,&cTimer1,4);                // Current Timer
   write (handle,&Timer1,4);                 // timer init value
   write (handle,&TIMER1_LO,4);              // value written to LOW?
   write (handle,&cTimer2,4);
   write (handle,&Timer2,4);
   write (handle,&TIMER2_LO,4);
   write (handle,&cTimer3,4);
   write (handle,&Timer3,4);
   write (handle,&TIMER3_LO,4);

   write (handle,Memory,0x10000);   // Dump ALL off memory!
   close (handle);
}


//****************************************************************************
//
// Peek and Deek. get a byte or word from memory
//
//****************************************************************************
byte PeekC ( word address )
{
   return (Memory[address]);
}
word DeekC ( word address )
{        
   unsigned short temp2;
   
   temp2=(unsigned short) (Memory[address+1]*256);
   temp2+=Memory[address];
   return (temp2);
}

//****************************************************************************
//
// Poke and Doke
//
//****************************************************************************
void PokeC ( word address, byte data )
{
// uint8 *pMem;
// void  (*Routine)();
//
// if (MemoryWriteExe[address]!=0){
//    Routine=MemoryWriteTable[address];
//    Routine();
// } else {
//    pMem=(uint8*) MemoryWriteTable[address];
//    *pMem=data;
// }
//
// if (mem_break==1) {
//    if (address==brk_address) do_break=1;
// }
}

void DokeC ( word address, word data )
{
   Memory[address]=data & 0xff;
   Memory[address+1]=(data & 0xff00)>>8;
   if (mem_break==1) {
      if (address==brk_address) do_break=1;
   }
}




//****************************************************************************
//
// wait for a keypress
//
//****************************************************************************
void waitkey()
{
   char key=0x0;

   while (key==0x0) {
      key=kbhit();
   }
}

//****************************************************************************
//
// 
//
//****************************************************************************
int32 GetHex(char *st)
{
   int i=0,i2,i3;
   
	st++; // past $
   while (st[0]!=0x0) {
      if ((st[0]>='a') && (st[0]<='f')) st[0]-=32;
      i=i*16;
      i3=0;
      for (i2=0;i2<16;i2++){
         if (st[0]==hexbuff[i2]) { i3=i2; }
      }
      i=i+i3;
      st++;
   }
   return (i);
}


//****************************************************************************
//
// 
//
//****************************************************************************
int32 Val(char *st)
{
   int i=0; 

   if (st[0]=='$') {
      i=GetHex(st);
   } else {
      while (st[0]!=0x0) {
         i=i*10;
         i=i+(st[0]-48);
         st++;
      }
   }
   return (i);
}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//************************************************************

uint32 GetSize (char *pName)
{
struct stat    STAT;
      uint32   size;

   stat(pName,&STAT);

   return (STAT.st_size);
}

//****************************************************************************
//
// Load data into memory
//
//****************************************************************************
void  LoadData ( uint8 *pMemory, int address, char *name )
{
   int handle;
   uint32   Size;

   Size=GetSize(name);
   
   handle=open (name,O_RDONLY|O_BINARY);
   if (handle!=-1) {
      //printf ("         Loading file\n");
      read (handle,pMemory+address,Size);
      close (handle);
   }
}


//****************************************************************************
//
// Load data into memory
//
//****************************************************************************
void  LoadGame (char *name)
{
   int handle;
   uint16   add;
   uint32   Size;
	char		Name[1024];

	sprintf (&Name[0],"%s%s",pFileName,name);
   Size=GetSize(&Name[0]);

   handle=open (&Name[0],O_RDONLY|O_BINARY);
   if (handle!=-1) {
      //printf ("         Loading Game\n");
      read (handle,&add,2);
      read (handle,&(Memory[add]),Size-2);
      close (handle);
   }
   Doke(0x2b,0x1001);                     // setup start of basic
   Doke(0x2d,(Size+add-2));
   Doke(0x9d,(Size+add-2));
}


//****************************************************************************
//
//  Hex - return Hex dig
//
//****************************************************************************
void Hex (char *buff,int number,int count)
{
   if (count==2){
      buff[0]=hexbuff[(number&0xF0)>>4];
      buff[1]=hexbuff[number&0x0F];
   } else {
      buff[0]=hexbuff[(number&0xF000)>>12];
      buff[1]=hexbuff[(number&0xF00)>>8];
      buff[2]=hexbuff[(number&0xF0)>>4];
      buff[3]=hexbuff[number&0xF];
   }
}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

//
// blue  = 1
// green = 2
// red   = 3
// 
void DumpMemory (int address)
{
   char  line[80];
   char  add[5]="0000";
   char  hx1[3]="00";
   char  hx2[3]="00";
   char  hx3[3]="00";
   char  hx4[3]="00";
   char  hx5[3]="00";
   char  hx6[3]="00";
   char  hx7[3]="00";
   char  hx8[3]="00";
   char  hx9[3]="00";
   char  hx10[3]="00";
   char  hx11[3]="00";
   char  hx12[3]="00";
   char  hx13[3]="00";
   char  hx14[3]="00";
   char  hx15[3]="00";
   char  hx16[3]="00";
	char	ASCII[9]="        ";
   uint32   lp;   
	uint32	row,col;
	uint32	Addresses[8];


  for (lp=0;lp<8;lp++)  {
		Addresses[lp]=address&0xffff;
      Hex(&add,address,4);
      Hex(&hx1,(int) (ASCII[0]=Peek((address++)&0xffff)),2);
      Hex(&hx2,(int) (ASCII[1]=Peek((address++)&0xffff)),2);
      Hex(&hx3,(int) (ASCII[2]=Peek((address++)&0xffff)),2);
      Hex(&hx4,(int) (ASCII[3]=Peek((address++)&0xffff)),2);
      Hex(&hx5,(int) (ASCII[4]=Peek((address++)&0xffff)),2);
      Hex(&hx6,(int) (ASCII[5]=Peek((address++)&0xffff)),2);
      Hex(&hx7,(int) (ASCII[6]=Peek((address++)&0xffff)),2);
      Hex(&hx8,(int) (ASCII[7]=Peek((address++)&0xffff)),2);

		ASCII[0]=C_Table[ASCII[0]];
		ASCII[1]=C_Table[ASCII[1]];
		ASCII[2]=C_Table[ASCII[2]];
		ASCII[3]=C_Table[ASCII[3]];
		ASCII[4]=C_Table[ASCII[4]];
		ASCII[5]=C_Table[ASCII[5]];
		ASCII[6]=C_Table[ASCII[6]];
		ASCII[7]=C_Table[ASCII[7]];
      sprintf (line,"b01t0e%s b01t0F %s %s %s %s %s %s %s %s  b01t0a%s",add,hx1,hx2,hx3,hx4,hx5,hx6,hx7,hx8,ASCII);
      PrintLine (41,16+lp,line);
  }
  PrintLine (51,12,"                             ");
  PrintLine (51,13,"                             ");


  // MARK edit bit.
  address-=64;
  row=0;
  col=0;
  for (lp=0;lp<64;lp++){
		if (EditMem==address) {
         Hex(&hx1,(int) (ASCII[0]=Peek(address)) ,2);
         sprintf (line,"b02t0E%cb01%c",hx1[0],hx1[1]);
			if (EditNibble==1){
				line[3]='1';
				line[12]='2';
			}
         PrintLine (41+6+(col*3),16+row,line);

			break;
		} else{
			col++;
			if (col==8) { col=0; row++; }
			address=(address+1)&0xffff;
		}
  }
}


//****************************************************************************
//
// Display Monitor Screen
//
//****************************************************************************
void Regs ( void )
{
  char   line[128];
  char   f[7];
  char   pc[5];
  char   ar[3];
  char   xr[3];
  char   yr[3];
  char   sp[3];
  int32  lp,i;    //,aa;
  char	VALUE[50];
  char	VALUE2[50];

//  return;

   strcpy (f,"0000000");
   strcpy (pc,"0000");
  	strcpy (VALUE,"0000");
  	strcpy (VALUE2,"00");
   strcpy (ar,"00");
   strcpy (xr,"00");
   strcpy (yr,"00");
   strcpy (sp,"00");
   Hex(&pc,_PC,4);
   Hex(&ar,_A,2);
   Hex(&xr,_X,2);
   Hex(&yr,_Y,2);
   Hex(&sp,_SP,2);

   if ((_F&_C)!=0){ f[6]='1'; }
   if ((_F&_Z)!=0){ f[5]='1'; }
   if ((_F&_I)!=0){ f[4]='1'; }
   if ((_F&_D)!=0){ f[3]='1'; }
   if ((_F&_B)!=0){ f[2]='1'; }
   if ((_F&_V)!=0){ f[1]='1'; }
   if ((_F&_N)!=0){ f[0]='1'; }


   for (i=0;i<80;i++) line[i]=0x0;
   sprintf (line,"b01t0a Addr b01t0ePeekb01t0f PC   AC  XR  YR  SP  NVBDIZC",_A,_X,_Y,_SP);
   PrintLine (41,0,line);
 
   for (i=0;i<80;i++) line[i]=0x0;
   sprintf (line,"b01t0a      b01t0e    b01t0f%s  %s  %s  %s  %s  %s",pc,ar,xr,yr,sp,f);
   PrintLine (41,1,line);
 
   //
   // make sure _PC falls at a Disassembled line, and if not make 
   // either SCROLL to include it, or SET the current disassmebler point.
   //
	
	aa=0;
   for (lp=0;lp<10;lp++) {
	   if (_PC==(StartDisPC+aa)) break;
 	   Disassemble(StartDisPC+aa);
      aa+=adder;
   }
   if (lp==10) {								// allow to SCROLL
	   if ((StartDisPC+aa)==_PC){
 		   Disassemble(StartDisPC);
		   StartDisPC+=adder;
	   }else{
  		   StartDisPC=_PC;
	   }
   } else{
   if ((StartDisPC+aa)!=_PC) StartDisPC=_PC;
   }




	//
   // now PRINT it.
	//
   aa=0;
   for (lp=0;lp<10;lp++) {
		//PrintSourceLine(lp, StartDisPC+aa);

      Disassemble(StartDisPC+aa);
      PrintLine (41,2+lp,"b01t00      ");
      PrintLine (47,2+lp,"b01t0e    ");

		if (Location!=0xffffffff){
 		   Hex(&VALUE,Location,4);
		   sprintf (line,"b01t0a\$%s",VALUE);
         PrintLine (41,2+lp,line);
		}

		if (LocationValue!=0xffffffff){
 		  	Hex(&VALUE2,LocationValue,2);
		   sprintf (line,"b01t0e\$%s",VALUE2);
         PrintLine (47,2+lp,line);
		}
	

	   if (_PC==(StartDisPC+aa)){
      	PrintLine (50,2+lp,"b02t0F                              ");
		   sprintf (line,"b02t0F>%s",disline);
         PrintLine (50,2+lp,line);
	   }else if ((BreakPoints[(StartDisPC+aa)]&BREAK_PC)!=0){
     	   PrintLine (50,2+lp,"b04t07                              ");
	      sprintf (line,"b04t0F %s",disline);
         PrintLine (50,2+lp,line);
		}else{
     	   PrintLine (50,2+lp,"b01t07                              ");
	      sprintf (line,"b01t07 %s",disline);
      	PrintLine (50,2+lp,line);
	   }
      aa+=adder;
   }	
   EndDisPC=_PC+aa;

	SourceLevelDis();


   DumpMemory (MemDump);



   PrintLine (41,12,"b01t0F                                       ");
   PrintLine (41,13,"b01t0F                                       ");
   PrintLine (41,14,"b01t0F                                       ");
   PrintLine (41,15,"b01t0F                                       ");
   
	// INFO
	sprintf (line,"b01t0FLineCycles=%d ",LineCycles);
   PrintLine (41,13,line);
   sprintf (line,"b01t0FRaster=%d ",(Peek(0xff1d)+(( ( Peek(0xff1c)&1 )*256) ) ));
   PrintLine (56,13,line);
   sprintf (line,"b01t0FTape=%d  Timer1=%x Sample=%d",SampleStart,cTimer1,SB_ready);
   PrintLine (41,14,line);

   
   // Print HELP line at the bottom
	PrintLine (0,24,"b04t0eMt07-Dump t0eBt07-BreakPoint t0eGt07-GOTO Addr t0eEntert07-Step t0eF9t07-Step Fast t0eEt07-Edit t0eRt07-ROM/RAM t0eF1t07-Help");

}


//****************************************************************************
//
// Various routines to get data from the Memory
//
//****************************************************************************
unsigned char GetByteC ( void )
{
   unsigned char temp;
   
   temp=Memory[_PC];
   _PC=(_PC+1)&0xffff;
   return (temp);
}

unsigned short GetWordC ( void )
{
   unsigned short temp2;
   
   temp2=(unsigned short) (Memory[_PC+1]*256);
   temp2+=Memory[_PC];
   _PC=(_PC+2)&0xffff;
   return (temp2);
}

//****************************************************************************
//
// Push and Dush
//
//****************************************************************************
void Push_C ( byte data )
{
   Memory[0x100+_SP]=data;
   _SP--;
}
void Dush_C ( word data )
{
   Memory[0x100+_SP-1]=data & 0xff;
   Memory[0x100+_SP]=(data & 0xff00)>>8;
   _SP-=2;
}

//****************************************************************************
//                      
// Pull and Dull
//
//****************************************************************************
void Pull ( byte *data )
{
   _SP++;
   *data=Memory[0x100+_SP];
}
void Dull ( word *data )
{
   _SP+=2;
   *data=Memory[0x100+_SP-1];
   *data+=(Memory[0x100+_SP]<<8);
}


//****************************************************************************
//
// Take a BRANCH instruction - signed offset byte is passed
//
//****************************************************************************
void TakeBranch( int8 offset)
{
   if (offset>127) {
      offset=-offset;
      _PC=(_PC-offset)&0xffff;
   } else {
      _PC=(_PC+(int) offset)&0xffff;
   }
}

//****************************************************************************
//
// Do the actual cmp command
//
//****************************************************************************
void DoCMP ( byte test )
{
   _F&=(_I+_D+_V+_B);
   if (test<128) {_F|=_N;}
   if (test==0) {_F|=_Z;}
   if (test<128) {_F|=_C;}
}

//****************************************************************************
//
// Do a compare with 0 (negative and zero checks)
//
//****************************************************************************
void COMP_C ( byte test )
{
   _F&=(_I+_D+_V+_C+_B);
   if ((test & 128)==128) {_F|=_N;}
   if (test==0) {_F|=_Z;}
}

#include "6502\adc.c"
#include "6502\alr.c"
#include "6502\and.c"
#include "6502\anc.c"
#include "6502\asl.c"
#include "6502\aso.c"
#include "6502\bcc.c"
#include "6502\bcs.c"
#include "6502\beq.c"
#include "6502\bit.c"
#include "6502\bmi.c"
#include "6502\bne.c"
#include "6502\bpl.c"
#include "6502\brk.c"
#include "6502\bvc.c"
#include "6502\bvs.c"
#include "6502\clc.c"
#include "6502\cld.c"
#include "6502\cli.c"
#include "6502\clv.c"
#include "6502\dcm.c"
#include "6502\cmp.c"
#include "6502\cpx.c"
#include "6502\cpy.c"
#include "6502\dec.c"
#include "6502\dex.c"
#include "6502\dey.c"
#include "6502\eor.c"
#include "6502\inc.c"
#include "6502\inx.c"
#include "6502\iny.c"
#include "6502\ins.c"
#include "6502\jmp.c"
#include "6502\jsr.c"
#include "6502\lda.c"
#include "6502\axs.c"
#include "6502\lax.c"
#include "6502\ldx.c"
#include "6502\ldy.c"
#include "6502\lse.c"
#include "6502\lsr.c"
#include "6502\nop.c"
#include "6502\ora.c"
#include "6502\oal.c"
#include "6502\pha.c"
#include "6502\php.c"
#include "6502\pla.c"
#include "6502\plp.c"
#include "6502\rla.c"
#include "6502\rol.c"
#include "6502\ror.c"
#include "6502\rra.c"
#include "6502\rti.c"
#include "6502\rts.c"
#include "6502\sax.c"
#include "6502\say.c"
#include "6502\sbc.c"
#include "6502\sec.c"
#include "6502\sed.c"
#include "6502\sei.c"
#include "6502\sta.c"
#include "6502\stx.c"
#include "6502\sty.c"
#include "6502\tax.c"
#include "6502\tay.c"
#include "6502\tsx.c"
#include "6502\txa.c"
#include "6502\txs.c"
#include "6502\tya.c"
#include "6502\skb.c"
#include "6502\skw.c"
#include "6502\hlt.c"
#include "6502\xaa.c"
#include "6502\jump.c"


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

int32 CountFF1e (char *pTable, uint8 *pFF1E)
{
	int	i,index;
	uint8	count;
	
	count=0xc4;								// start of SCAN line
	index=0;
 	for (i=0;i<114;i++){
		if (pTable[i]=='p'){
			pFF1E[index]=count;
			index++;
		}
		count+=2;
		if (count>=0xe4) count-=0xe4;
	}
	return	index;
}
void InitFF1e ( void)
{
	cfastdma=CountFF1E (&FastDMA[0],&tfastdma[0]);
	cslowdma=CountFF1E (&SlowDMA[0],&tslowdma[0]);
	cfastnormal=CountFF1E (&FastNormal[0],&tfastnormal[0]);
	cslownormal=CountFF1E (&SlowNormal[0],&tslownormal[0]);
	cfastboarder=CountFF1E (&FastBoarder[0],&tfastboarder[0]);
	cslowboarder=CountFF1E (&SlowBoarder[0],&tslowboarder[0]);
}

//************************************************************
//                                                           *
//   Name    : InitTedSamples											 *
//   Function: Build a lookup table for the TED sample       *
//             so FF11 can be used DIRECTLY as an index.     *
//                                                           *
//************************************************************

void InitTedSamples ( void )
{
	//uint8	TEDSample[256];
	float	TempSample,tone;
	int	i;

	for (i=0;i<256;i++){
		if ( ((i&0x80)!=0) AND ((i&0x30)!=0) ){
         TempSample=((float) (i&0xf));
         if (TempSample>8) TempSample=8.0f;
         TempSample=TempSample*0.125;					// /8.0f;
         if ((i&0x30)==0x30){
				tone=255.0f*TempSample;					// *2 if BOTH channels are playing
			} else {
         	tone=127.5f*TempSample;						// 0 to 128 for 1 channel playing
			}
		} else {
			tone=0;												// 0 for NO channels playing.
		}
		TEDSample[i]=(uint8) tone;
	}
}


//****************************************************************************
//
// Init memory. Allocate etc..
//
//****************************************************************************
void  init6502( void )

{
   uint32   i,i2,i3,i4;
   uint32   *pMem;
   uint32   *pROM;

	FF1e_DONELINE=0;
	FASTLOADER=0;
	TapeHighEdge=TapeHighEdgeTable[TapeEdge];
	TapeLowEdge=TapeLowEdgeTable[TapeEdge];


   pTimer1Lo=(uint8 *) (&cTimer1);
   pTimer1Lo+=2;
   pTimer1Hi=(uint8 *) (&cTimer1);
   pTimer1Hi+=3;

   FlashCount=0;     // Cursor flash counter

   printf ("         Allocating Plus4 System Memory\n");
   pMemory=MyMalloc (0x10000+0x2000);
   if (pMemory==NULL) {
      printf ("\nError: Can't allocate system memory.\n");
      exit(1);
   }
   Memory=pMemory+0x1fff;
   Memory=(uint8 *) ((uint32) Memory&0xffffe000);

   printf ("         Allocating 1Mb for the memory mapping system\n");
   MemoryWrite=(uint32 *) MyMalloc (0x40000);                  // a mapping system for data access
   MemoryRead=(uint32 *) MyMalloc (0x40000);                // a mapping system for data access (what to store)
   MemoryRead2=(uint32 *) MyMalloc (0x40000);                  // a mapping system for data access (what to store)
   if (MemoryWrite==NULL) {
      printf ("\nError: Can't Allocate Memory for Memory Mapping System.\n");
      exit(1);
   }
   if (MemoryRead==NULL) {
      printf ("\nError: Can't Allocate Memory for Memory Mapping System.\n");
      exit(1);
   }
   if (MemoryRead2==NULL) {
      printf ("\nError: Can't Allocate Memory for Memory Mapping System.\n");
      exit(1);
   }

   printf ("        Clearing memory and tables\n");
   for (i=0;i<0x10000;i++) {
      Memory[i]=0;                           // clear memory
      MemoryRead[i]=((uint32) &Memory[i])&0xfffffffc;    // Set default mode (read from memory!)
      MemoryRead2[i]=((uint32) &Memory[i])&0xfffffffc;
      MemoryWrite[i]=((uint32) &Memory[i])&0xfffffffc;      // Set default mode (write to memory!)
		BreakPoints[i]=0;								// clear break point table
   }
   MemoryWriteTable=MemoryWrite;
   MemoryReadTable=MemoryRead;

   // ROM on/off
   MemoryWrite[0xff3e]=(uint32) &RomOn|ROUTINE;          // ROM switched off!
   MemoryWrite[0xff3f]=(uint32) &RomOff|ROUTINE;            // ROM switched on!

   // 
   // Set up ROM numbers. Used in ROM paging
   //
   MemoryWrite[0xfdd0]=(uint32) &ROMINF_400|ROUTINE;     // MemoryWrite[0xfdd0]=0x400;                   
   MemoryWrite[0xfdd1]=(uint32) &ROMINF_401|ROUTINE;     // MemoryWrite[0xfdd1]=0x401;                  
   MemoryWrite[0xfdd2]=(uint32) &ROMINF_402|ROUTINE;     // MemoryWrite[0xfdd2]=0x402;                  
   MemoryWrite[0xfdd3]=(uint32) &ROMINF_403|ROUTINE;     // MemoryWrite[0xfdd3]=0x403;                  
   MemoryWrite[0xfdd4]=(uint32) &ROMINF_500|ROUTINE;     // MemoryWrite[0xfdd4]=0x500;            
   MemoryWrite[0xfdd5]=(uint32) &ROMINF_501|ROUTINE;     // MemoryWrite[0xfdd5]=0x501;            
   MemoryWrite[0xfdd6]=(uint32) &ROMINF_502|ROUTINE;     // MemoryWrite[0xfdd6]=0x502;
   MemoryWrite[0xfdd7]=(uint32) &ROMINF_503|ROUTINE;     // MemoryWrite[0xfdd7]=0x503;
   MemoryWrite[0xfdd8]=(uint32) &ROMINF_600|ROUTINE;     // MemoryWrite[0xfdd8]=0x600;            
   MemoryWrite[0xfdd9]=(uint32) &ROMINF_601|ROUTINE;     // MemoryWrite[0xfdd9]=0x601;            
   MemoryWrite[0xfdda]=(uint32) &ROMINF_602|ROUTINE;     // MemoryWrite[0xfdda]=0x602;            
   MemoryWrite[0xfddb]=(uint32) &ROMINF_603|ROUTINE;     // MemoryWrite[0xfddb]=0x603;            
   MemoryWrite[0xfddc]=(uint32) &ROMINF_700|ROUTINE;     // MemoryWrite[0xfddc]=0x700;            
   MemoryWrite[0xfddd]=(uint32) &ROMINF_701|ROUTINE;     // MemoryWrite[0xfddd]=0x701;            
   MemoryWrite[0xfdde]=(uint32) &ROMINF_702|ROUTINE;     // MemoryWrite[0xfdde]=0x702;            
   MemoryWrite[0xfddf]=(uint32) &ROMINF_703|ROUTINE;     // MemoryWrite[0xfddf]=0x703;                   
                                           
   MemoryWrite[0xff06]=(uint32) &ProcessFF06|ROUTINE;
   MemoryWrite[0xfef0]=(uint32) &DiskDMAByte|ROUTINE;
   MemoryRead[0xfef1]=(uint32) &DiskEND|ROUTINE;
   MemoryRead2[0xfef1]=(uint32) &DiskEND|ROUTINE;
   MemoryWrite[0x01]=(uint32) &SerialBits|ROUTINE;


   MemoryWrite[0xff1f]=(uint32) &_FF1F|ROUTINE;
                                      
   MemoryRead[0xff1e]=(uint32) &_FF1E|ROUTINE;
   MemoryRead2[0xff1e]=(uint32) &_FF1E|ROUTINE;

	if (FASTLOADER==0){
	   MemoryWrite[0x032e]=(uint32) &_32e|ROUTINE;		// don't allow writing here!
	   MemoryWrite[0x032f]=(uint32) &_32f|ROUTINE;
	}


   AddSound();
   // 
   MemoryWrite[0xff1d]   =(uint32) &_FF1D|ROUTINE;


   // flag keyboard writing.
   MemoryWrite[0xff08] =(uint32) &wKEYBOARD_LATCH|ROUTINE;     
   MemoryWrite[0xfd30] =(uint32) &wKEYBOARD_LATCH_2|ROUTINE;   
   MemoryRead[0xff08]=(uint32) &rKEYBOARD_LATCH|ROUTINE;
   MemoryRead2[0xff08]=(uint32) &rKEYBOARD_LATCH|ROUTINE;

   // IRQ wiping
   MemoryWrite[0xff09] =(uint32) &Clear_IRQS|ROUTINE;

   // Setting the timer          
   MemoryWrite[0xff00]=(uint32) &wTIMER1_LO|ROUTINE;           
   MemoryWrite[0xff01]=(uint32) &wTIMER1_HIGH|ROUTINE;         
   MemoryWrite[0xff02]=(uint32) &wTIMER2_LO|ROUTINE;           
   MemoryWrite[0xff03]=(uint32) &wTIMER2_HIGH|ROUTINE;
   MemoryWrite[0xff04]=(uint32) &wTIMER3_LO|ROUTINE;           
   MemoryWrite[0xff05]=(uint32) &wTIMER3_HIGH|ROUTINE;


	// Temp write in SID sound stuff
   MemoryWrite[0xfd40]=(uint32) &_fd40|ROUTINE;
   MemoryWrite[0xfd41]=(uint32) &_fd41|ROUTINE;
   MemoryWrite[0xfd42]=(uint32) &_fd42|ROUTINE;
   MemoryWrite[0xfd43]=(uint32) &_fd43|ROUTINE;
   MemoryWrite[0xfd44]=(uint32) &_fd44|ROUTINE;
   MemoryWrite[0xfd45]=(uint32) &_fd45|ROUTINE;
   MemoryWrite[0xfd46]=(uint32) &_fd46|ROUTINE;

   MemoryWrite[0xfd47]=(uint32) &_fd47|ROUTINE;
   MemoryWrite[0xfd48]=(uint32) &_fd48|ROUTINE;
   MemoryWrite[0xfd49]=(uint32) &_fd49|ROUTINE;
   MemoryWrite[0xfd4a]=(uint32) &_fd4a|ROUTINE;
   MemoryWrite[0xfd4b]=(uint32) &_fd4b|ROUTINE;
   MemoryWrite[0xfd4c]=(uint32) &_fd4c|ROUTINE;
   MemoryWrite[0xfd4d]=(uint32) &_fd4d|ROUTINE;

   MemoryWrite[0xfd4e]=(uint32) &_fd4e|ROUTINE;
   MemoryWrite[0xfd4f]=(uint32) &_fd4f|ROUTINE;
   MemoryWrite[0xfd50]=(uint32) &_fd50|ROUTINE;
   MemoryWrite[0xfd51]=(uint32) &_fd51|ROUTINE;
   MemoryWrite[0xfd52]=(uint32) &_fd52|ROUTINE;
   MemoryWrite[0xfd53]=(uint32) &_fd53|ROUTINE;
   MemoryWrite[0xfd54]=(uint32) &_fd54|ROUTINE;

   MemoryWrite[0xfd55]=(uint32) &_fd55|ROUTINE;
   MemoryWrite[0xfd56]=(uint32) &_fd56|ROUTINE;
   MemoryWrite[0xfd57]=(uint32) &_fd57|ROUTINE;
   MemoryWrite[0xfd58]=(uint32) &_fd58|ROUTINE;

	//-----
   MemoryWrite[0xfe80]=(uint32) &_fd40|ROUTINE;		// SID mirror
   MemoryWrite[0xfe81]=(uint32) &_fd41|ROUTINE;
   MemoryWrite[0xfe82]=(uint32) &_fd42|ROUTINE;
   MemoryWrite[0xfe83]=(uint32) &_fd43|ROUTINE;
   MemoryWrite[0xfe84]=(uint32) &_fd44|ROUTINE;
   MemoryWrite[0xfe85]=(uint32) &_fd45|ROUTINE;
   MemoryWrite[0xfe86]=(uint32) &_fd46|ROUTINE;

   MemoryWrite[0xfe87]=(uint32) &_fd47|ROUTINE;
   MemoryWrite[0xfe88]=(uint32) &_fd48|ROUTINE;
   MemoryWrite[0xfe89]=(uint32) &_fd49|ROUTINE;
   MemoryWrite[0xfe8a]=(uint32) &_fd4a|ROUTINE;
   MemoryWrite[0xfe8b]=(uint32) &_fd4b|ROUTINE;
   MemoryWrite[0xfe8c]=(uint32) &_fd4c|ROUTINE;
   MemoryWrite[0xfe8d]=(uint32) &_fd4d|ROUTINE;

   MemoryWrite[0xfe8e]=(uint32) &_fd4e|ROUTINE;
   MemoryWrite[0xfe8f]=(uint32) &_fd4f|ROUTINE;
   MemoryWrite[0xfe90]=(uint32) &_fd50|ROUTINE;
   MemoryWrite[0xfe91]=(uint32) &_fd51|ROUTINE;
   MemoryWrite[0xfe92]=(uint32) &_fd52|ROUTINE;
   MemoryWrite[0xfe93]=(uint32) &_fd53|ROUTINE;
   MemoryWrite[0xfe94]=(uint32) &_fd54|ROUTINE;

   MemoryWrite[0xfe95]=(uint32) &_fd55|ROUTINE;
   MemoryWrite[0xfe96]=(uint32) &_fd56|ROUTINE;
   MemoryWrite[0xfe97]=(uint32) &_fd57|ROUTINE;
   MemoryWrite[0xfe98]=(uint32) &_fd58|ROUTINE;


   printf ("         Allocating ROM space\n");
   for (i=0;i<MAXROM;i++){
      pROMS[i]=MyMalloc (0x4000+0x2000);
      ROMS[i]=pROMS[i]+0x1fff;;
      if (ROMS[i]==NULL) {
         printf ("\nError: Can't Allocate Memory ROM's.\n");
         exit(1);
      }
      ROMS[i]=(uint8 *) ((uint32)ROMS[i] &0xffffe000);   // allocate 16k rom space.
      pMem=(uint32*) ROMS[i];
      for (i2=0;i2<0x1000;i2++) *pMem++=0;
   }


   printf ("         Loading basic ROM\n");
   LoadData(ROMS[0],0x0000,"c16basic.rom");     // load ROM to 0!
   printf ("         Loading kernal ROM\n");
   LoadData(ROMS[4],0x0000,"c16kern.rom");      // load ROM to 0

   LoadData(ROMS[1],0x0000,"3plus1lo.rom");
   LoadData(ROMS[5],0x0000,"3plus1hi.rom");
   *(ROMS[4]+0x3f08)=0xff;
   *(ROMS[5]+0x3f08)=0xff;
   *(ROMS[6]+0x3f08)=0xff;
   *(ROMS[7]+0x3f08)=0xff;

   printf ("         Loading character ROM\n");
   //LoadData(ROMS[4],0x1000,"c16char.rom");    


   // copy PAGE routine into ROM's($fc00-$fcff)
   pROM=(uint32*) ROMS[4];
   for (i2=5;i2<MAXROM;i2++){
      pMem=(uint32*) ROMS[i2];
      for (i=0xf00;i<0xf00+64;i++){
         pMem[i]=pROM[i];
      }
   }


   mem_break=0;
   ROM=1;               // Set roms active!
   ROMINF=0x400;        // Set default roms to use

   //
   // Now, set up memory WRITEs to go to rom LOW and HIGH routines
   //
   SetRomZone (0x400);
   
   _PC=Deek(0xfffc);    // 0xf2a4;        // reset computer
   _SP=0xFF;
   _A=0;
   _Y=0;
   _X=0;
   _F=0;

   _A=0xff;

   ScreenLine=0;
   RasterLine=SCREEN_LINES-20;
   Memory [0xff1c]=((RasterLine>>8)|0xfe);      // set continuous raster
   Memory [0xff1d]=(RasterLine&255);            // set continuous raster
	Memory [0xff09]=0x4; 								// LightPen
	pCurrentFF1E=&tfastboarder[0];
	LineCycles=0;
	CYCLES_PER_LINE=109;									// fast boarder=109
	Memory[0xff1e]=0xd0;
	ff1e_old=Memory[0xff1e];
	
	Memory[0xff10]=0x7c;

   pScreenData=&ScreenData[0];
   pScreenDataAddress=&ScreenData[0];
   _screen=0;
   
   i3=0;
   for (i=0;i<25;i++){
     for (i2=0;i2<40;i2++) {
      ScreenIndexArray[(i*8*40)+i2]=i3;
      ScreenIndexArray[(i*8*40)+i2+40]=i3;
      ScreenIndexArray[(i*8*40)+i2+(40*2)]=i3;
      ScreenIndexArray[(i*8*40)+i2+(40*3)]=i3;
      ScreenIndexArray[(i*8*40)+i2+(40*4)]=i3;
      ScreenIndexArray[(i*8*40)+i2+(40*5)]=i3;
      ScreenIndexArray[(i*8*40)+i2+(40*6)]=i3;
      ScreenIndexArray[(i*8*40)+i2+(40*7)]=i3;
      i3++;
     }
   }
   ProcessJoyStick ();
   ScanKeyBoard();                     // read keyboard for this frame.

	InitTedSamples();
	InitFF1E();
}


//****************************************************************************
//                                                           -
//   Name    :                                               -
//   Function:                                               -
//                                                           -
//   In   :                                                  -
//   Out  :                                                  -
//                                                           -
// This need rewritten. Its very confusing. i.e. BAD! 
//
// uint8    CharAttribute[40];            // read on the LAST line of the previous character
// uint8    CharNumber[40];               // read on the FIRST line of THIS character
//
//****************************************************************************

ProcessScanLine(void)
{
   byte     opcode,op2,data,a;
   char     key;
   int      I;
   uint16   rast,charset,charscreen,i;
   uint32   curs;
   uint16   Rast;
   uint32   Line;
   int32    Offset=0;
   int32    dontdraw=0,CL;
   uint8    *index;
   int32		StartCharLine[8]={0,0,0,0,0,0,0,0};
   int32		StartCharLine2[8]={0,0,0,0,-40,-40,-40,-40};

   RasterLine=Memory[0xff1d]+((Memory[0xff1c]&1)<<8);
	pScreenData=&ScreenData[0];
   //Poke (0xff1e,0xff);

	if (RasterLine==4){
		ScreenBlank = (Memory[0xff06]&0x10);      // screen blanked?
	}
   if (ScreenLine<244){
      pScreenData->blank   = ScreenBlank;	//(Memory[0xff06]&0x10);      // screen blanked?

      //
      // get most of the data needed to render the scanline.
      //
      pScreenData->multicol= (Memory[0xff07]&0x10)|(Memory[0xff06]&0x40);  //MCM & ECM
      pScreenData->Invert  = (Memory[0xff07]&0x80);
      pScreenData->ScrollX = (Memory[0xff07]&0x0f);
      pScreenData->ScrollY = (Memory[0xff06]&0x0f);
      pScreenData->col0 = (Memory[0xff15]&0x7f);         // background colour
      pScreenData->col1 = (Memory[0xff16]&0x7f);
      pScreenData->col2 = (Memory[0xff17]&0x7f);
      pScreenData->col3 = (Memory[0xff18]&0x7f);
      pScreenData->col4 = (Memory[0xff19]&0x7f);         // border colour

      scroffset=4;      //(pScreenData->ScrollY&7)+1;


      //
      // Now check the Bit Map Mode (or BMM)
      //
      pScreenData->bmm  = (Memory[0xff06]&0x20);
      if (pScreenData->bmm!=0){
         pScreenData->BitMap =((Memory[0xff12]&0x38)<<10);     // move bits to the TOP!
      }


      //
      // work out screen address
      //
      charscreen=Memory[0xff14]&(~7);
      if ((ROM==0) OR (charscreen<0x80)) {
         pScreenData->ScreenAddress=(uint32)((charscreen<<8)+Memory);
      } else {
         if (charscreen>=0xc0){
            pScreenData->ScreenAddress=(uint32) (((charscreen<<8)-0xc000)+ROMS[(ROMINF>>8)&0xff]);
         } else {
            pScreenData->ScreenAddress=(uint32) (((charscreen<<8)-0x8000)+ROMS[(ROMINF&0xff)]);
         }
      }


      //
      // get Cursor address
      //
      curs = (Memory[0xff0C]&3)*256;
      curs = Memory[0xff0D]+curs;
      pScreenData->Cursor = curs;


     
		CharLine=Memory[0xff1f]&7;
      CharRow=(Memory[0xff1a]<<8)+Memory[0xff1b];
      if (RasterLine==0){  //((pScreenData->ScrollY+1)&7)) {
		   
			y1=Memory[0xff06]&7;
			CharRow=0;                            // NEXT character ROW has been reached.
			if (y1==7) CharRow=40;
         index=(uint8 *)pScreenData->ScreenAddress+CharRow;
			for (i=0;i<40;i++){
            CharNumber[i]=index[0x400];   // read NEXT character ROW
            CharAttribute[i]=*index;      // read NEXT Attribute ROW
				index++;
			}

			CharRow=-40;
			if (y1==7) CharRow=-80;
			y1=(7-y1-1)&7;
         CharLine=y1; //StartCharLine[y1];
      }

		if (skipit==0){
			CharLine++;
			CharLine&=0x07;

			if (CharLine==0){
				CharRow+=40;                            // NEXT character ROW has been reached.
				CharRow&=0x3ff;
				index=(uint8 *)pScreenData->ScreenAddress+CharRow;
				for (i=0;i<40;i++){
					CharNumber[i]=index[0x400];   // read NEXT character ROW
					CharAttribute[i]=*index;      // read NEXT Attribute ROW
					index++;
				}
			}
		}
		skipit=0;

		Offset=0;
      Memory[0xff1f]=CharLine|((FlashCount&0xf)<<3)|0x80;	//+0xf0;
      Memory[0xff1a]=(CharRow>>8)&3;
      Memory[0xff1b]=(CharRow&255);



      //
      // work out character set..
      // 
		if (Memory[0xff07]&0x80){
      	charset=Memory[0xff13]&(~7);
		}else{
	      charset=Memory[0xff13]&(~3);
		}


      if (((Memory[0xff12]&4)==0) OR (charset<0x80)){
         pScreenData->CharSet=(uint32) (((charset<<8)+Memory)+(CharLine&7));
      } else {
         if (charset>=0xc0){
            pScreenData->CharSet=(uint32) (((charset<<8)-0xc000)+ROMS[(ROMINF>>8)&0xff]+(CharLine&7));
         } else {
            pScreenData->CharSet=(uint32) (((charset<<8)-0x8000)+ROMS[(ROMINF&0xff)]+(CharLine&7));
         }
      }

      if (Offset!=0){
         if (((ScreenLine-scroffset-Offset)>=20) AND 
             ((ScreenLine-scroffset-Offset)<=23)) {
            dontdraw=1;
         }
      } 

      if ((RasterLine>=(200+scroffset)) OR (RasterLine<scroffset)) {
         pScreenData->blank   = 0;     // blank screen
      } 

      //
      // Y scroll active! (well borders active)
      //
      if (pScreenData->blank!=0){
         if ((pScreenData->ScrollY&8)==0){
            if ((RasterLine-Offset>=4) AND (RasterLine-Offset<=7)){
               dontdraw=1;
            } else if ((RasterLine-Offset>=200) AND (RasterLine-Offset<=203)){
               dontdraw=1;
            } 
         }
      }
      //
      // DRAW scan line!
      //
      charcount++;
      if (pScreenData->blank!=0){
         if ((pScreenData->ScrollY&8)==0){
            if (ScreenLine-scroffset-Offset<=20) 
            pScreenData->blank   = 0;     // blank screen
         }
      }
        
      if (pScreenData->blank==0){
         Offset=0;
      }

      

      if ( (ScreenLine-scroffset-Offset)>0){
         if (dontdraw==0){
            PrintScreen((ScreenLine-scroffset-Offset)*Xsize,
                        (RasterLine-scroffset)*40,
                        ((RasterLine-scroffset)&7)+(((RasterLine-scroffset)>>3)*320));
         }
      }

      //
      // Fix up borders
      //
      if ((pScreenData->ScrollY&8)==0){
         if ((RasterLine>=4) AND (RasterLine<=7)){
               pScreenData->blank   = 0;     // blank screen
               PrintScreen((ScreenLine-scroffset)*Xsize, 0,    0);
         } else if ((RasterLine>=200) AND (RasterLine<=204)){
               pScreenData->blank   = 0;     // blank screen
               PrintScreen((ScreenLine-scroffset)*Xsize, 0,    0);
         } 
      }
   }

}

void pause(void)
{
}

//****************************************************************************
//
// Process a 6502 Instruction
//
//****************************************************************************
void Process6502 ( void )
{
   byte     op2,data,vol;
   char     key;
   int      I,i;
   uint16   rast;
   uint8    T;
   uint16   tim;
   uint8    CHAR;
   void     *test;
   uint16   TT;
   uint32   TimerCounter,time;
   int32    tempirq;
   float    TempSample,tone;
   uint8    *pSampleBuffer;
	int8		SAMPLE;	
	uint8		CharLine;


   //
   // Process actual 6502 instruction.
   //
   opcode=GETABYTE();
   Opcode_JumpTable[opcode]();
	_F|=0x30;

	// It seems that simple things like memory[address]=Value takes a
	// long time in C... so theres an assembler routine to do it
	// all faster...
	DoBits();								// do some basic stuff in assembler...
	ff1e_old=Memory[0xff1e];


   // *************************************************
   //
   // Raster is "incremented" at FF1E=$BC (376/2)
	//
	// *************************************************
//	if (Memory[0xff1e]>=0xbc){
//		if (ResetRasterLine==0){
//         RasterLine=Memory[0xff1d]+((Memory[0xff1c]&1)<<8);
//			RasterLine++;
//         
//		   if (RasterLine>SCREEN_LINES){
//            RasterLine=0;
//		   }
//         Memory [0xff1c]=((RasterLine>>8)|0xfe);      // set continuous raster
//         Memory [0xff1d]=(RasterLine&255);            // set continuous raster
//			ResetRasterLine=1;
//		}
//	}else{
//		ResetRasterLine=0;
//	}
//
//   IRQrast=Memory[0xff0b]+((Memory[0xff0a]&1)<<8);    // get raster IRQ line
//	if (RasterLine==IRQrast){
//		pause();
//	}

   //
   // Do END raster things - $FF1E = $D0, end of HBLANK
   //
   if (LineCycles>=CYCLES_PER_LINE) {
      LineCycles-=CYCLES_PER_LINE;        					// next line count
	//if (DoLine==0){
      DoLine=0;
      IRQrast=Memory[0xff0b]+((Memory[0xff0a]&1)<<8);    // get raster IRQ line
      ScreenLine++;

      RasterLine=Memory[0xff1d]+((Memory[0xff1c]&1)<<8);
      if (ScreenLine>SCREEN_LINES) {
			int i;
         ScreenLine=0;
         RasterLine=SCREEN_LINES-21;
	      Memory [0xff1c]=((RasterLine>>8)|0xfe);      // set continuous raster
   	   Memory [0xff1d]=(RasterLine&255);            // set continuous raster
      }

      // Restart frame
      if (ScreenLine==0){
         FlashCount++;
			if ((Memory[0xff07]&0x80)!=0){
   	      FlashCount=(FlashCount&0x0f)|0x10;
			}else{
   	      FlashCount=FlashCount&0x1f;
			}
         ProcessJoyStick ();
         ScanKeyBoard();                     // read keyboard for this frame.
         ScreenOff=1;
         DoCursor();
         DoMouse ();
         DoWindowStuff(0x0);
			//CopySample(pBuffer1);
         CopyScreen(1);    // refresh..
			//for (i=0;i<368*240;i++){
			//	Screen[i]=0;
			//}
      }


      //
      // If raster was forced by user, don't INC
      //
      if (DontInc==0) {
			RasterLine++; 
		}else{
			DontInc=0;
		}

      // normal...
		if (RasterLine>SCREEN_LINES){
         RasterLine=0;	//SCREEN_LINES;
	   }
      Memory [0xff1c]=((RasterLine>>8)|0xfe);      // set continuous raster
      Memory [0xff1d]=(RasterLine&255);            // set continuous raster


      // *************************************************
      //
      // Work out sample value for START of raster
      // pBuffer1 is the TED sample buffer.
		//
      // *************************************************
		pBuffer1[ScreenLine*2]=TEDSample[Memory[0xff11]];
		if (EnableNoise==1){
			data=pNoise[NoiseIndex];
			vol=Memory[0xff11]&0xf;
			if (vol>8) vol=8;
			if (vol==0)
				data=0;
			else{
				vol=(8-vol)+1;
				data=data/vol;
			}
			pBuffer2[ScreenLine*2]=data;
			NoiseStep+=NoiseStepDelta;
			NoiseIndex+=(NoiseStep>>16);
			NoiseIndex&=0xffff;
			NoiseStep&=0xffff;
		}else{
			pBuffer2[ScreenLine*2]=0;
		}

		ProcessSID(ScreenLine*2);
	}

	//
	// Calculate MID raster line sample values.
	//
	if (LineCycles>=(CYCLES_PER_LINE>>1)) {
		if (MidSample==0){
			pBuffer1[(ScreenLine*2)+1]=TEDSample[Memory[0xff11]];
			MidSample=1;

			if (EnableNoise==1){
				data=pNoise[NoiseIndex];
			   vol=Memory[0xff11]&0xf;
			   if (vol>8) vol=8;
			   if (vol==0)
				   data=0;
			   else{
				   vol=(8-vol)+1;
				   data=data/vol;
			   }
			   pBuffer2[(ScreenLine<<SoundShifter)]=data;
			   NoiseStep+=NoiseStepDelta;
			   NoiseIndex+=(NoiseStep>>16);
			   NoiseIndex&=0xffff;
			   NoiseStep&=0xffff;
			}else{
				pBuffer2[(ScreenLine*2)+1]=0;
			}
			ProcessSID((ScreenLine*2)+1);
		}
	}else{
		MidSample=0;
	}

   // *************************************************
   //
   // Process RASTER interupt
   //
   // *************************************************
//   if (( (IRQrast==RasterLine) OR ((Memory[0xff09]&2)!=0)) AND (ThisLine==0)){
//      if ( (LineCycles>IRQ_START_CYCLE) ){
//      	Memory[0xff09]|=2;         				// set IRQ NOW!!!
//		   ThisLine=1;									// flag don't process THIS line until NEXT rasterline
//		}
//   } else if (IRQrast!=RasterLine) ThisLine=0;


   // ***************************************************
   //
	// Now work out next lines cycle usage, and new $FF1E
   //
   // ***************************************************
	if (DoLine==0){
		//DoLine=0;
      DoLine=1;

	   CharLine=Memory[0xff1f]&7;
	   if (skipit==0){
	      CharLine++;
	      CharLine&=7;
	   }else{
	      CharLine=1;
	   }

	   if ((Memory[0xff13]&2)!=0){
	      //CPU_MODE=_1MHZ;
            	
	      // First, is screen off or in upper or lower boarder area?
	      if (((Memory[0xff06]&0x10)==0) OR ((RasterLine<3) OR (RasterLine>203))){
	         CYCLES_PER_LINE=cslowboarder;				// slow boarder=57
		      pCurrentFF1E=&tslowboarder[0];

	      // Or, are we on a slow "DMA" line (0 or 7)?
	      }else if ((CharLine==0) OR (CharLine==7)){
	         CYCLES_PER_LINE=cslowdma;					// slow DMA=14 (+3 "write" cycles)
		      pCurrentFF1E=&tslowdma[0];

	      } else {
	         CYCLES_PER_LINE=cslownormal;					// slow normal=57
		      pCurrentFF1E=&tslownormal[0];
	
	      }
	   } else{
	      //CPU_MODE=_1_78MHZ;

	      // First, is screen off or in upper or lower boarder area?
	      if (((Memory[0xff06]&0x10)==0) OR ((RasterLine<3) OR (RasterLine>203))){
		      CYCLES_PER_LINE=cfastboarder;					// fast boarder=109
		      pCurrentFF1E=&tfastboarder[0];

	      }else if ((CharLine==0) OR (CharLine==7)){
		      CYCLES_PER_LINE=cfastdma;					// fast DMA=22 (+3 "write" cycles)
		      pCurrentFF1E=&tfastdma[0];
	      } else {
		      CYCLES_PER_LINE=cfastnormal;					// fast normal=65
		      pCurrentFF1E=&tfastnormal[0];
	      }
	   }
	   IRQ_START_CYCLE=8;	

      //
		// This is only used by the TAPE loading system now
      //
      mem_ff1e_delta=0xE40000/CYCLES_PER_LINE;        // work out delta for scan line
      CycleIndex[0]=0;
      for (i=1;i<25;i++){
         CycleIndex[i]=CycleIndex[i-1]+mem_ff1e_delta;      // work out deltas for EVERY instuction length
      }
      mem_ff1e=0;    //CycleIndex[LineCycles];              // Add on the number of cycles used.
	}
	Memory[0xff1e]=pCurrentFF1E[LineCycles];




	// Inc raster line at correct point
//	if ( (Memory[0xff1e]>0xbc) AND (FF1e_DONELINE==0)) {
//		FF1e_DONELINE=1;
//
//      RasterLine=Memory[0xff1d]+((Memory[0xff1c]&1)<<8);
//		RasterLine++;
//		if (RasterLine>SCREEN_LINES) RasterLine=0;
//
//      IRQrast=Memory[0xff0b]+((Memory[0xff0a]&1)<<8);    // get raster IRQ line
//		if (IRQrast==RasterLine){
//	         Memory[0xff09]|=0x02;               // set IRQ NOW!!!
//		}
//
//	   Memory [0xff1c]=((RasterLine>>8)|0xfe);      // set continuous raster
//   	Memory [0xff1d]=(RasterLine&255);            // set continuous raster
//	}else{
//		if (Memory[0xff1e]<0xbc) {
//			if (FF1e_DONELINE==2) {
//				FF1e_DONELINE=3;
//			}
//		}
//	}
//
//   // *************************************************
//   //
//   // Process RASTER interupt, also check for end of HBlank
//   //
//   // *************************************************
//	if (FF1e_DONELINE==1) {
//		if ((Memory[0xff1e]>0xce) OR (Memory[0xff1e]<0x40)) {
//			FF1e_DONELINE=2;
//
//		   if ((Memory[0xff0a]&Memory[0xff09]&0x02)!=0) {	      // Gen IRQ for this timer
//	   	   Memory[0xff09]|=0x80;               // set IRQ NOW!!!
//		   }
//		}
//	}

// Working...
   if (( (IRQrast==RasterLine) OR ((Memory[0xff09]&2)!=0)) AND (ThisLine==0)){
      if ( (LineCycles>IRQ_START_CYCLE) ){
      	Memory[0xff09]|=2;         				// set IRQ NOW!!!
		   ThisLine=1;										// flag don't process THIS line until NEXT rasterline
		}
		if ((Memory[0xff0a]&Memory[0xff09]&0x02)!=0) {	      // Gen IRQ for this timer
			Memory[0xff09]|=0x80;               // set IRQ NOW!!!
		}
   } else if (IRQrast!=RasterLine) ThisLine=0;


//	if (Memory[0xff1e]<0xbc){
//   	if (  (IRQrast==RasterLine) OR ((Memory[0xff09]&2)!=0) AND (ThisLine==0) ){
//      	Memory[0xff09]|=2;         				// set IRQ NOW!!!
//		   ThisLine=1;										// flag don't process THIS line until NEXT rasterline
//		}
//		if ((Memory[0xff0a]&Memory[0xff09]&0x02)!=0) {	      // Gen IRQ for this timer
//			Memory[0xff09]|=0x80;               // set IRQ NOW!!!
//		}
//   } else if (IRQrast!=RasterLine) ThisLine=0;



   // ***************************************************
   //
	// Process timers 1,2 and 3
   //
   // ***************************************************

	if (ff1e_old>Memory[0xff1e]){			// looped?
		time=0xe4-ff1e_old+Memory[0xff1e];
	}else{
		time=Memory[0xff1e]-ff1e_old;
	}
   if (TIMER1_LO==0){                     // process counter?
   	TimerCounter=cTimer1;                  // get current timer
		cTimer1-=(time<<14);

      if (cTimer1>TimerCounter) {
         cTimer1+=Timer1;                 // reset timer
			if ((Memory[0xff0a]&0x08)!=0){	// Gen IRQ for this timer
	         Memory[0xff09]|=0x88;         // set IRQ NOW!!!
			}else{
	         Memory[0xff09]|=0x08;         // Just set flag, but don't gen IRQ
			}
		}
      Memory[0xff00]=*pTimer1Lo;    // write NEW timer LOW
      Memory[0xff01]=*pTimer1Hi;    // write NEW timer HIGH
	}

	// Process timer 2
   if (TIMER2_LO==0){                     	// process counter?
   	TimerCounter=cTimer2;                  // get current timer
		cTimer2-=(time<<14);

      if (cTimer2>TimerCounter) {
			if ((Memory[0xff0a]&0x10)!=0){	// Gen IRQ for this timer
	         Memory[0xff09]|=0x90;         // set IRQ NOW!!!
			}else{
	         Memory[0xff09]|=0x10;         // Just set flag, but don't gen IRQ
			}
		}
      Memory[0xff02]=(cTimer2>>16)&0xff;    // write NEW timer LOW
      Memory[0xff03]=(cTimer2>>24)&0xff;    // write NEW timer HIGH
   }

	// Process timer 3
   if (TIMER3_LO==0){                     	// process counter?
   	TimerCounter=cTimer3;                  // get current timer
		cTimer3-=(time<<14);

      if (cTimer3>TimerCounter) {
			if ((Memory[0xff0a]&0x40)!=0){	// Gen IRQ for this timer
	         Memory[0xff09]|=0xc0;         // set IRQ NOW!!!
			}else{
	         Memory[0xff09]|=0x40;         // Just set flag, but don't gen IRQ
			}
		}
      Memory[0xff04]=(cTimer3>>16)&0xff;    // write NEW timer LOW
      Memory[0xff05]=(cTimer3>>24)&0xff;    // write NEW timer HIGH
   }



	//
	// ANY interupts happened? Timer 1,2,3 OR raster?
	//
   //if (Memory[0xff09]&Memory[0xff0a]&0x5a){
   if ((Memory[0xff09]&0x80)!=0){ 			//Memory[0xff0a]&0x5a){
      data=Memory[0xff0a];                // get IRQ switch
      if ((_F&_I)==0) {               		// IRQ's ON??
			DushReg (_PC);
		   PushReg (_F&0xef);
         _F|=_I;                          // disable irqs...
         _PC=Deek(0xfffe);                // get IRQ vector
         NextCycles=7;                    // 7 for the START of an IRQ
      }
	}

   // *************************************************
   //
	// Process TAPE sample
   //
   // *************************************************
	if (SampleStart==1){								// Has "virtual" play been pressed?
		Memory[0xfd10]=(Memory[0xfd10]&~4);		// PLAY pressed


	   if (mem_ff1e_old>mem_ff1e){
	   	TapeSoFar+=((mem_ff1e+0xE30000)-mem_ff1e_old);
	   } else {
	   	TapeSoFar+=(mem_ff1e-mem_ff1e_old);
		}

		if (TapeSoFar>=TapeCountCheck){
			if ((Memory[1]&8)==0){
				if (*pHeader2>=TapeHighEdge){
				   Memory[1]=((Memory[1]&~0x10));	
		  	   }else if (*pHeader2<=TapeLowEdge){
				   Memory[1]=((Memory[1]|0x10));	
			   }
			   pHeader2++;
				if (pHeader2>=(pHeader+TapeFileSize)) {
					FASTLOADER=0;
					SampleStart=0;												// stop tape!
					Memory[0xfd10]=(Memory[0xfd10]|4);					// press STOP
	      		MemoryWrite[0x032e]=(uint32) &_32e|ROUTINE;		// don't allow writing here!
	      		MemoryWrite[0x032f]=(uint32) &_32f|ROUTINE;		// (enable this again)
					MyFree(pHeader);
					pHeader=NULL;
				}
			}
			TapeSoFar-=TapeCountCheck;
		}
	}

	

   // *************************************************
   //
	// Draw this actual scan line (current one) $FF1E=0
   //
   // *************************************************

   if (LineCycles>10){
	//if (FF1e_DONELINE==3) {
		FF1e_DONELINE=0;
		if (DRAWIT==0)
		{
			DRAWIT=1;
         ProcessScanLine();

         if ((ModeX!=0) AND (ScreenLine<244) AND (debug==0)){
            Screen2=(ScreenLine*368)+(uint32)Screen;
            VidBase2=Frame+(ScreenLine*(368/4));
				if (DOCOPYSCREEN==0){
            	CopyXmodeLine ();
				}
         }
		}
	} else {
		DRAWIT=0;
	}



}




//****************************************************************************
//                                                           -
//   Name    :                                               -
//   Function:                                               -
//                                                           -
//****************************************************************************

void wipe ( void )
{
   int i;
   for (i=0;i<256;i++){
      Keys2[i]=0;
   }
}

void wipe2 ( void )
{
   int i;
   for (i=0;i<256;i++){
      Keys[i]=0;
   }
}

//****************************************************************************
//
// Debug
//
//****************************************************************************
void Process ( void )
{
   byte     op2;
   char     key;
   int      I,skip,address,i,handle,x,y;
   char     s[128];
   word     t;
   uint8    FF1D;
	uint8	   DoneSomething,	Attribute;
	uint8	   nibble,m;
	uint32	Memtmp;
	char		*pAdd1,*pAdd2,*pFileName,*pVal;
	int32		StartAddress,EndAddress;
  	char   	line[1024],newline;
	uint16	LoadAddress,Shift;

   key=0x0;
   
   _ret=0;
   key=0x0;
   skip=0;
	for (i=0;i<65536;i++){
		BreakPoints[i]&=~SKIP_PC;
	}
   
   wipe();  
	DoneSomething=1;
   while ((_ret==0)) {
		if (DoneSomething){
			Regs();
			DoneSomething=0;
		}
      //
      // HELP
      //
      if (Keys2[K_F1]) {
			wipe();
			Attribute=TextColour;
			SetPrintCol(0x1f);					// Blue background, Yellow Text
			i=OpenTextWindow (10,0,60,24," Help - Escape to continue ");
			
			x=15;
			y=1;
			PrintLine (x,y,   "q        t0e- Quit and return to DOS");
			PrintLine (x,y+1, "ESCAPE   t0e- Return to emulator");
			PrintLine (x,y+2, "m        t0e- Set memory dump address");
			PrintLine (x,y+3, "e        t0e- Edit memory");
			PrintLine (x,y+4, "U/D/L/R  t0e- Scroll Memory Window");
			PrintLine (x,y+5, "PgU/PgD  t0e- Page Memory Window");
			PrintLine (x,y+6, "g        t0e- Set PC address");
			PrintLine (x,y+7, "b        t0e- Set a breakpoint at address");
			PrintLine (x,y+8, "r        t0e- Toggle ROM/RAM in/out");
			PrintLine (x,y+9, "s        t0e- Save memory");
			PrintLine (x,y+10,"S        t0e- Save .PRG");
			PrintLine (x,y+11,"p        t0e- Freeze+Pack to disk");
			PrintLine (x,y+12,"ENTER    t0e- trace into (slow)");
			PrintLine (x,y+13,"F9       t0e- trace into (fast)");
			PrintLine (x,y+14,"SPACE    t0e- step over (branches+jsr\'s)");
			PrintLine (x,y+15,"F12      t0e- Warm reset");
			PrintLine (x,y+16,"Shift+F12t0e- Cold reset");
			PrintLine (x,y+17,"F6       t0e- Save BMP (snapshot of current screen)");
			PrintLine (x,y+18,"F5       t0e- Enter BITMAP trace mode");
			PrintLine (x,y+19,"           ENTER t0e- Trace a frame");
			PrintLine (x,y+20,"           SPACE t0e- Trace a scanline");
			PrintLine (x,y+21,"           ESCAPEt0e- Return to debugger");
		
			while ((Keys2[K_ENTER]==0) AND (Keys2[K_ESCAPE]==0));
			
			CloseTextWindow(i);

			SetPrintCol(Attribute);		
         wipe();  
      }


      //
      // RUN
      //
      if (Keys2[K_ESCAPE]) {
         wipe();  
			wipe2();
         return;
      }

      if (Keys2[K_F6])  {              // save PCX file
			pFileName=Input(10,10,60," Enter BMP filename ",57);		// input string
			if (pFileName!=NULL){
            SaveDIB(pFileName);
			   MyFree(pFileName);
			}
			DoneSomething=1;
      }

      if (Keys2[K_F5])  {              // VIEW screen (so far)
         wipe();  
         ScreenMode=BitMapScreen;
         InitScreen ();
         _screen=1;
         CopyScreen (1);
         while (Keys2[K_ESCAPE]==0) {              // VIEW screen (so far)
            if (Keys2[K_SPACE]){
               wipe();
               FF1D=Memory[0xff1d];
               while (FF1D==Memory[0xff1d]){
                  Process6502();       // trace..
               }
               CopyScreen (1);
           }
          if (Keys2[K_ENTER]){
               wipe();
               // run one line,
               FF1D=Memory[0xff1d];
               while (FF1D==Memory[0xff1d]){
                  Process6502();       // trace..
               }
               // then run a SCREEN worth
               while ( (Memory[0xff1d]+((Memory[0xff1c]&1)<<8)) !=0 ){
                  Process6502();       // trace..
               }
               CopyScreen (1);
           }
         }
         SetMode(3);
         ScreenMode=CharScreen;
         _screen=1;
         CopyScreen (1);
         wipe();
			DoneSomething=1;
      }

		// **********************************************
		//
		// EDIT memory in HEX
		//
		// **********************************************

      if (Keys2[K_E])  {              // EDIT memory
			EditMem=MemDump;
			Regs();
         wipe();  
         while (Keys2[K_ESCAPE]==0) {              // keep edit until ESCAPE
			//for(i=0;i<128;i++){
			//	if (Keys2[i]!=0){
         //    	printf("%d\n",i);
         //			wipe();  
			//			break;
			//	}
			//}
				nibble=0xff;
            if (Keys2[K_0]) nibble=0;
            if (Keys2[K_1]) nibble=1;
            if (Keys2[K_2]) nibble=2;
            if (Keys2[K_3]) nibble=3;
            if (Keys2[K_4]) nibble=4;
            if (Keys2[K_5]) nibble=5;
            if (Keys2[K_6]) nibble=6;
            if (Keys2[K_7]) nibble=7;
            if (Keys2[K_8]) nibble=8;
            if (Keys2[K_9]) nibble=9;
            if (Keys2[K_A]) nibble=10;
            if (Keys2[K_B]) nibble=11;
            if (Keys2[K_C]) nibble=12;
            if (Keys2[K_D]) nibble=13;
            if (Keys2[K_E]) nibble=14;
            if (Keys2[K_F]) nibble=15;

				if (nibble!=0xff){
				   m=Peek(EditMem);
				   if (EditNibble==0){
					   m=(m&0x0f)|(nibble<<4);
					   Poke(EditMem,m);
					   EditNibble++;
				   }else{
					   m=(m&0xf0)|nibble;
					   Poke(EditMem,m);
					   EditNibble=0;
					   EditMem=(EditMem+1)&0xffff;
				   }
					if (EditMem==((MemDump+64)&0xffff)){		// scroll down
						MemDump=((MemDump+8)&0xffff);
					}
		         DoneSomething=1;
				}

				// scroll edit point
            if (Keys2[K_UP]) {
		         EditMem=(EditMem-8)&0xffff;
		         DoneSomething=1;
		      }
            if (Keys2[K_DOWN]) {
		         EditMem=(EditMem+8)&0xffff;
		         DoneSomething=1;
		      }
            if (Keys2[K_LEFT]) {
					if (EditNibble==1)
						EditNibble=0;
					else{
						EditNibble=1;
			         EditMem=(EditMem-1)&0xffff;
					}
		         DoneSomething=1;
		      }
            if (Keys2[K_RIGHT]) {
					if (EditNibble==0)
						EditNibble=1;
					else{
						EditNibble=0;
			         EditMem=(EditMem+1)&0xffff;
					}
		         DoneSomething=1;
		      }

				Memtmp=(MemDump-8)&0xffff;
				for (i=0;i<8;i++){
					if ((Memtmp+i)==EditMem) MemDump=(MemDump-8)&0xffff;
				}
				Memtmp=(MemDump+64)&0xffff;
				for (i=0;i<8;i++){
					if ((Memtmp+i)==EditMem) MemDump=(MemDump+8)&0xffff;
				}

            if (Keys2[K_PAGEUP]) {
			      EditMem=(EditMem-64)&0xffff;
			      MemDump=(MemDump-64)&0xffff;
		         DoneSomething=1;
				}
            if (Keys2[K_PAGEDOWN]) {
			         EditMem=(EditMem+64)&0xffff;
			         MemDump=(MemDump+64)&0xffff;
		         DoneSomething=1;
				}
         	
			
				// refresh edit window (if we did something)
				if (DoneSomething==1){
					Regs();
					DoneSomething=0;
            	wipe();
				}
         }
			EditMem=0x10000;
         wipe();
			DoneSomething=1;
      }

		// **********************************************
		// End Edit Memory	
		// **********************************************


      //
      // dissassemble to a file
      //
      if (Keys2[K_D]) {
			wipe();
			pAdd1=Input(30,10,20," Enter Start Addr ",5);			// input string
			pAdd2=Input(30,10,20," Enter Stop Addr ",5);			// input string
			pFileName=Input(10,10,60," Enter FileName ",57);		// input string

			if ((pAdd1!=0x0) AND (pAdd2!=0x0) AND (pFileName!=0x0)){
				ClearCallTable();
				StartAddress=Val(pAdd1);
				EndAddress=Val(pAdd2);
				RangeLow=StartAddress;
				RangeHigh=EndAddress;

				// PASS1 - Get Calls and branches so we can make labels
      		while(StartAddress<EndAddress){
      			Disassemble(StartAddress);
               StartAddress+=adder;
				}


				StartAddress=RangeLow;
				handle=open (pFileName,O_WRONLY|O_BINARY|O_CREAT,S_IRWXU|S_IRWXG);
				newline=0xd;
      		while(StartAddress<EndAddress){

      			Disassemble(StartAddress);
		   		sprintf (line,"L%s%c\n",disline,newline);
					if (CallTable[StartAddress]==0){
						line[0]=0x20;	line[1]=0x20;	line[2]=0x20;	
						line[3]=0x20; 	line[4]=0x20;
					}
					if (CallBranchAddress!=0xffffffff){
					   if (CallTable[CallBranchAddress]!=0){
								line[14]='L';
					   }
					}
					write (handle,&line[0],Len(line));
               StartAddress+=adder;
				}
				close(handle);
			}

			MyFree(pFileName);
			MyFree(pAdd1);
			MyFree(pAdd2);
         wipe();  
			DoneSomething=1;
      }

      //
      // Save memory to a file, or save PRG
      //
      if (Keys2[K_S]) {
			if ((Keys[K_RIGHTSHIFT]) OR (Keys[K_LEFTSHIFT]))
				Shift=1;
			else
				Shift=0;
			
			pAdd1=Input(30,10,20," Enter Start Addr ",5);			// input string
			pAdd2=Input(30,10,20," Enter End Addr ",5);			// input string
			pFileName=Input(10,10,60," Enter FileName ",57);		// input string

			if ((pAdd1!=0x0) AND (pAdd2!=0x0) AND (pFileName!=0x0)){
				StartAddress=Val(pAdd1);
				EndAddress=Val(pAdd2);
				handle=open (pFileName,O_WRONLY|O_BINARY|O_CREAT,S_IRWXU|S_IRWXG);
				LoadAddress=(uint16)StartAddress;
				if (Shift==1){
					write (handle,&LoadAddress,2);
				}
				write (handle,&Memory[StartAddress],(EndAddress-StartAddress)+1);
				close(handle);
			}

			MyFree(pFileName);
			MyFree(pAdd1);
			MyFree(pAdd2);
         wipe();  
			DoneSomething=1;
      }

      //
      // fill memory
      //
      if (Keys2[K_F]) {
			pAdd1=Input(30,10,20," Enter Start Addr ",5);			// input string
			pAdd2=Input(30,10,20," Enter End Addr ",5);			// input string
			pVal=Input(10,10,60," Enter FILL value ",57);		// input string

			if ((pAdd1!=0x0) AND (pAdd2!=0x0) AND (pVal!=0x0)){
				StartAddress=Val(pAdd1);
				EndAddress=Val(pAdd2);
				m=Val(pVal);
				for (i=StartAddress;i<(EndAddress+1);i++){
					Memory[i]=m;
				}
			}
			MyFree(pVal);
			MyFree(pAdd1);
			MyFree(pAdd2);
		}

      //
      // Restart
      //
      if (Keys2[K_F12]) {
         FIndex=0;
         RomOn();             // ROM=1 - Set roms active!
         SetRomZone(0x400);   // Set defualt roms!
         DushReg (_PC);
         PushReg (_F);
         _PC=Deek(0xfffc);    // 0xf2a4;        // reset computer
         
         Keys[K_F12]=0;
         wipe();  
			wipe2();
         return;
      }


      if (Keys2[K_B]){ 
			pAdd1=Input(22,10,36," Enter Break Point Address ",5);			// input string
			if (pAdd1!=NULL){
            newpc=Val(pAdd1);          
			   MyFree(pAdd1);
			}
			BreakPoints[newpc]^=BREAK_PC;
         wipe();
			DoneSomething=1;
		} 

      //
      // run until address is "touched"
      //
//         if (Keys2[K_B]) { 
//         wipe();   
//         Restore_Keyboard ();
//         scanf("%s",s);
//         if (s[0]=='$')  
//         brk_address=GetHex(&s);
//         else        
//         brk_address=Val(&s);          
//         mem_break=1; 
//         Initialise_Keyboard ();
//         wipe();
//		   DoneSomething=1;
//         return;
//		}

      //
      // dump memory
      //
      if(Keys2[K_M]) { 
			pAdd1=Input(22,10,36," Enter DUMP start addresss ",5);			// input string
			if (pAdd1!=NULL){
            MemDump=Val(pAdd1);          
			   MyFree(pAdd1);
			}
         skip=1;   
         wipe();
			DoneSomething=1;
		}


      if (Keys2[K_Q]) { EXIT_PRG=1; skip=2; wipe(); wipe2(); return; }

      if (Keys2[K_R]){
         wipe();   
			if (ROM==0) 
				RomOn();
			else
				RomOff();
			DoneSomething=1;
		}
      //
      // goto address (PC=?)
      //
      if (Keys2[K_G]){
			pAdd1=Input(22,10,36," Enter new PC address ",5);			// input string
			if (pAdd1!=NULL){
            _PC=Val(pAdd1);          
			   MyFree(pAdd1);
			}
         skip=1;   
         wipe();
			DoneSomething=1;
		}


		// Scroll Memory Window around (fast+slow)
      if (Keys2[K_UP]) {
		   wipe();
		   MemDump=(MemDump-8)&0xffff;
		   DoneSomething=1;
		}
      if (Keys2[K_DOWN]) {
		   wipe();
		   MemDump=(MemDump+8)&0xffff;
		   DoneSomething=1;
		}
      if (Keys2[K_LEFT]) {
		   wipe();
		   MemDump=(MemDump-1)&0xffff;
		   DoneSomething=1;
		}
      if (Keys2[K_RIGHT]) {
		   wipe();
		   MemDump=(MemDump+1)&0xffff;
		   DoneSomething=1;
		}
		if (Keys2[K_PAGEUP]) {
		   wipe();
		   MemDump=(MemDump-64)&0xffff;
		   DoneSomething=1;
		}
		if (Keys2[K_PAGEDOWN]) {
		   wipe();
			MemDump=(MemDump+64)&0xffff;
		   DoneSomething=1;
		}


      if (Keys2[K_ENTER]) {
         wipe();
         Process6502();       // trace..

         ScreenMode=CharScreen;
         _screen=1;
         CopyScreen (1);

			Keys2[K_ENTER]=0;
			DoneSomething=1;
      }

      if (Keys2[K_SPACE]) {
         wipe();
			opcode=GETABYTE(); _PC--;
			switch (opcode){
				case	0x90:
				case	0xb0:
				case	0xf0:
				case	0x30:
				case	0xd0:
				case	0x10:
				case	0x50:
				case	0x70: BreakPoints[_PC+2]|=SKIP_PC;
								wipe2();
								return;
				case	0x20: BreakPoints[_PC+3]|=SKIP_PC;
								wipe2();
								return;
				default:
         		Process6502();
			}
         ScreenMode=CharScreen;
         _screen=1;
         CopyScreen (1);

			DoneSomething=1;
      }

      if (Keys2[K_P]) {
			pFileName=Input(10,10,60," Enter 'freeze' filename ",57);		// input string
         
			if (pFileName!=0x0){
				Pack(pFileName);
			}
			MyFree(pFileName);
         wipe();
      }


      //if (Keys2[K_O]) {
		//	UnPack();
      //   wipe();
      //}

      if (Keys[K_F9]) {
         Process6502();       // trace..
         ScreenMode=CharScreen;
         _screen=1;
         CopyScreen (1);
			DoneSomething=1;
      }
      if (skip==2) return;
      skip=0;
   }
}

//****************************************************************************
//
// Process a 6502 Instruction
//
//****************************************************************************
void Process2 ( void )
{
   word     t;
   byte     op2,joytog=0,joytog2=0;
   char     key;
   int      I,skip,address,i;
   char     s[128];
   char     line[80];
   char     f[7];
   char     pc[5];
   char     ar[3];
   char     xr[3];
   char     yr[3];
   char     sp[3];
   int      lp,aa;
   uint32   tt;
   int   handle;
   uint16   Add,doneline=0;
	char		name[1024];

   key=0x0;
   
   _ret=0;
   key=0x0;
   skip=0;
   
   //Poke (0xff08,0xff);      // keyboard


   wipe();
   tt=0;
   while (1){
      Process6502();
      if (ScreenLine==10 && doneline==0) {
         
         doneline=1;
         if ((Keys[K_BREAK]!=0) AND (Keys[K_LEFTALT]!=0)) { 
            EXIT_PRG=1; 
            wipe();
            return; 
         }

         if (Keys[K_F10]) {
            if (pFileName==NULL){
               LoadProgram(".");
            } else {
               LoadProgram (pFileName);
            }
            Keys[K_F10]=0;
         }

         if (Keys[K_F11]) { 
            debug=1;
            Keys[K_F11]=0;
            return;
         }

         if (Keys[K_F7]) { 
            if (f7==0){
               if (EnableMouse==0){
						InitWindowSystem ();					  // open options window
                  //pMaster->Options&=~HIDE;         // enable window system
                  EnableMouse=1;
               } else {
		            CloseWindow(pMaster);
	               pMaster=NULL;
                  EnableMouse=0;
               }
            f7=1;
            }
         }  else {
            f7=0;
         }  

         if (Keys[K_F12]) { 
				for (i=0xfd40;i<0xfd5f;i++){		// stop SID sound
					Memory[i]=0;
				}
				FileLoading=0; 								// NO file loading now
				if (LoadingBar==0) CloseWindow (pLoading);
				pLoading=NULL;

            FIndex=0;
            if (Keys[K_LEFTSHIFT] OR Keys[K_RIGHTSHIFT]){
               RomOn();             // ROM=1 - Set roms active!
               SetRomZone(0x400);   // Set defualt roms!
            }
            DushReg (_PC);
            PushReg (_F);
            _PC=Deek(0xfffc);    // 0xf2a4;        // reset computer
            Keys[K_F12]=0;
         }

//       if (Keys[K_F5]) {
//          _A=_A+1;
//          _A=_A-1;
//          Keys[K_F5]=0;
//       }

         if (Keys[K_F9]) {
            FreezeGame();
            Keys[K_F9]=0;
         }

         if (Keys[K_F8]) {
		      if (pHeader!=NULL) MyFree(pHeader);							// tape in memory already?
    	      Keys[K_F8]=0;
		      SampleStart=1;
		      TapeCountCheck=(int32)(TAPE_COUNT*65536.0f);				// 0x504ca0 (pal)
		      TapeSoFar=0;
		      sprintf	(&name[0],"%s%s",pFileName,TapeFileName);
		      pHeader2=pHeader=(uint8*) WindLoad(&name[0]);
		      TapeFileSize=GetFileSize(&name[0]);
		      Memory[0xfd10]=(Memory[0xfd10]&~4);										// press PLAY
            MemoryWrite[0x032e]=((uint32) &Memory[0x032e])&0xfffffffc;      // enable fast loaders
            MemoryWrite[0x032f]=((uint32) &Memory[0x032f])&0xfffffffc;      
		      FASTLOADER=1;
         }

         //if (Keys[K_F5]) {
         //}


         if (Keys[K_F6]) {
    	      Keys[K_F6]=0;
		      ScaleScreen^=1;
         }

      } else {
         if (ScreenLine!=10) doneline=0;
      }

      if ((BreakPoints[_PC]&(BREAK_PC|SKIP_PC))!=0) {
         wipe();
         return;        // PC break point!
      }
      //if (do_break==1) return;    // TOUCHED memory location
   }
}







