//****************************************************************************
//
//
//        The Minus4 project - Copyright (c) Mike Dailly 1996,97,98
//
//
// This source and the Minus4 project remains the copyright of Mike Dailly, 
// but can be freely used and changed for NON-Commercial purposes. 
//
//								 Keep emulators free!
//
// Remember if you add or change anything, you MUST send a copy
// complete with source to me at mike@scotch.demon.co.uk
// If its just a bug fix, then just send the changed file, with a comment
// telling me WHERE you fixed and WHAT you fixed.
//
// 
//										 SOUNDB.C
//                                      
// This file holds the SOUND BLASTER stuff. This deals with playing plus/4
// samples (samples on the plus/4 are played by "clicking" the volume).
// It will also deal with the Plus/4 SID chip.
//
// Version  By   Date		Description
//  V0.01	MD	12/10/97		File started. play a sample. (test)
//  V0.10	MD	15/10/97		Sample playing for Minus/4 added
//  V1.00	MD	14/01/98		Autodetect the soundblaster
//                                 		 
//
//
//****************************************************************************
#include "headers\md.h"
#include "headers\all.h"
#include "headers\screen.h"
#include "headers\draw.h"
#include "headers\6502.h"
#include "headers\6502asm.h"
#include "headers\disasm.h"
#include "headers\keyb.h"
#include	"headers\keyboard.h"
#include "headers\wind.h"
#include <I86.h>
#include <dos.h>



// PIC MASKS FOR MASK/DEMASK IRQ
#define	PICANDMASK		0x7f			// 01111111b 'AND' PIC mask for clear IRQ7
#define	PICORMASK		0x80			// 10000000b 'OR' PIC mask for set IRQ7

// DMA CONTROLLER REGISTERS :
#define	WRITEMASK		0x0a			// WRITE MASK REGISTER
#define	WRITEMODE		0x0b			// WRITE MODE REGISTER
#define	CLEARFLIPFLOP	0x0c
#define	PAGE_CHN			0x83			// PAGE REGISTER FOR DMAchannel 1
#define	BASE_CHN			0x02			// BASEADDRESS REGISTER DMA 1
#define	COUNT_CHN		0x03			// COUNT REGISTER DMAchannel 1

#define DMA_BUF_SIZE    8192
#define DMA8_FF_REG      0xC
#define DMA8_MASK_REG    0xA
#define DMA8_MODE_REG    0xB
#define DMA16_FF_REG    0xD8
#define DMA16_MASK_REG  0xD4
#define DMA16_MODE_REG  0xD6

#define DMA0_ADDR        0
#define DMA0_COUNT       1
#define DMA0_PAGE     0x87
#define DMA1_ADDR        2
#define DMA1_COUNT       3
#define DMA1_PAGE     0x83
#define DMA3_ADDR        6
#define DMA3_COUNT       7
#define DMA3_PAGE     0x82
#define DMA5_ADDR     0xC4
#define DMA5_COUNT    0xC6
#define DMA5_PAGE     0x8B
#define DMA6_ADDR     0xC8
#define DMA6_COUNT    0xCA
#define DMA6_PAGE     0x89
#define DMA7_ADDR     0xCC
#define DMA7_COUNT    0xCE
#define DMA7_PAGE     0x8A

#define DSP_BLOCK_SIZE            0x0048
#define DSP_DATA_AVAIL               0xE
#define DSP_HALT_SINGLE_CYCLE_DMA 0x00D0
#define DSP_READ_PORT               0xA
#define DSP_READY                   0xAA
#define DSP_RESET                   0x6
#define DSP_TIME_CONSTANT         	0x0040
#define DSP_WRITE_PORT            	0xC
#define DSP_VERSION               	0xE1

#define AUTO_INIT                   1
#define FAIL                        0
#define FALSE                       0
#define MASTER_VOLUME            0x22
#define MIC_VOLUME               0x0A
#define MIXER_ADDR                0x4
#define MIXER_DATA                0x5
#define MONO                        0
#define PIC_END_OF_INT           0x20
#define PIC_MASK                 0x21
#define PIC_MODE                 0x20
#define SUCCESS                     1
#define SINGLE_CYCLE                0
#define STEREO                      1
//#define TRUE                        1
#define VOICE_VOLUME             0x04


// SAMPLERATE : (if you change it pay attention to maximum samplerate)
// TimeConst= 256-(1000000/SampleRateHz)
//#define	TIMECONST		165         // = 10989 Hz (256-1000000/11000)


// DMA WRITE MODE
//#define	WANTEDMODE		0x48			// 0x1001000b singlemode, nonautoinit, readmode
#define	WANTEDMODE		0x58			// 0x1001000b singlemode, autoinit, readmode


uint8 disable( void );
#pragma aux disable =		\
		"cli"				\

uint8 enable( void );
#pragma aux enable =		\
		"sti"				\

uint8 WAITWRITE( uint16 port );
#pragma aux WAITWRITE =		\
		"push	cx"				\
		"xor	cx,cx"			\
"lp:	dec	cx"				\
		"jz	endloop"			\
		"in	al,dx"			\
		"or	al,al"			\
		"js	lp"				\
"endloop:	pop	cx"		\
		parm caller [dx]		\
		modify [al]		\
	   value [al]


uint8 WAITREAD( uint16 port );
#pragma aux WAITREAD =		\
		"push	cx"				\
		"xor	cx,cx"			\
"lp:	dec	cx"				\
		"jz	endloop"			\
		"in	al,dx"			\
		"or	al,al"			\
		"jns	lp"				\
"endloop:	pop	cx"		\
		parm caller [dx]		\
		modify [al]		\
	   value [al]

typedef	struct tSB_DMA_Data{
	void  *SB_DMA_Cont; 		  	// pointer to continous allocated memory
	uint16 SB_DMA_Selector; 	// Base selector
	int32  SB_DMA_Channel;		// channel for dma transfer
	int32  SB_DMA_Mode; 		   // transfer mode
	uint16 SB_DMA_Length;		// # of bytes to transfer
	uint8	 SB_DMA_Page; 		   // page
}tSB_DMA_Data;

	uint32			SB_ON=0;					// SB "deactive" by default 
	uint32			SBNextFrameReady=0;
volatile	uint32			SB_ready=0;			// ready for NEXT sample. (last one has stopped)
	uint32			SBcount=0;
	void 				(__interrupt __far *Old_IRQ_Isr)();		// old IRQ vector
	tSB_DMA_Data	buffer1;											// DMA buffer (space for 2)
	uint8	*p,*p2;

	// have "8" sample buffers ready for mixing

	uint8		*pBuffer1;	// TED samples
	uint8		*pBuffer2;	// TED noise
	uint8		*pBuffer3;	// SID channel 1
	uint8		*pBuffer4;	// SID channel 2
	uint8		*pBuffer5;	// SID channel 3
	uint8		*pBuffer6;	// SID samples
	uint8		*pBuffer7;	// spare
	uint8		*pBuffer8;	// spare

	uint32	BufferList[2];
	uint32	DMAchannel=1;           // SoundBlaster DMA channel
	uint32	BaseAddr=0x220;			// SoundBlaster BASE address
	uint32	IRQ=7;						// SoundBlaster IRQ

unsigned int SB_IRQ_Mask1=-1;
unsigned int SB_IRQ_Mask2=-1;

	uint8 SB_DMA_Page[8]  ={0x87,0x83,0x81,0x82,0x8F,0x8B,0x89,0x8A};
	uint8	SB_DMA_Offset[8]={0x00,0x02,0x04,0x06,0xC0,0xC4,0xC8,0xCC};
	uint8	SB_DMA_Length[8]={0x01,0x03,0x05,0x07,0xC2,0xC6,0xCA,0xCE};

// Misc. DMA registers (8&16bit)
// -----------------------------
#define SB_DMA_CMASK8	0x0A		// 8 Bit DMA Channel Mask reg.
#define SB_DMA_CMASK16	0xD4		// 16 Bit DMA Channel Mask reg.
#define SB_DMA_CLEAR8	0x0C		// 8 Bit DMA Reset adress ptr reg.
#define SB_DMA_CLEAR16	0xD8		// 16 Bit DMA Reset adress ptr reg.
#define SB_DMA_MODE8		0x0B		// 8 Bit DMA Transfer Mode reg.
#define SB_DMA_MODE16	0xD6		// 16 Bit DMA Transfer Mode reg.


// This is 312 raster lines*50Hz refresh  = 15600
// This allows "amiga" style DMA, i.e. we get the data at the START of 
// each line. A simple way... Doubling it would be "nicer".
//	uint32	TIMECONST=192;         		// = 15600 Hz (256-1000000/11000)
//	uint32	BUFFER_SIZE=312;
	uint32	TIMECONST=224;         		// = 31200 Hz (256-1000000/11000)
	uint32	BUFFER_SIZE=624;

	int32		Base, DSP_Ver;
   uint8		ReD=0x80;

  int BufToFill,
      DMAChan8Bit,
      DMAChan16Bit,
      IRQNumber,
      IRQMask,
      MaskSave,
      Offset;

void interrupt DMAOutputISR(void);   // Interrupt Service Routine
  void interrupt (*IRQSave)(void);

	int32	PLAYSAMPLE=0xff;
	int32	PLAYSAMPLE2=0;
		uint8		*pSaw;		// SAWtooth sample base

//***********************************************************
//                                                          * 
//   Name    : SetColour                                    * 
//   Function: Sets a colour register to an R,G,B value     * 
//                                                          * 
//   In   : ColR=Colour Register to change                  * 
//          R   = Red   Colour Value                        * 
//          G   = Green Colour Value                        * 
//          B   = Blue  Colour Value                        * 
//   Out  : None                                            * 
//                                                          * 
//                                                          *
//***********************************************************

void	interSetColour (int ColR, int R, int G, int B)
{
	outp ( ColourNo,   (unsigned char) ColR );
	outp ( ColourAttr, (unsigned char) R);// >> 2);
	outp ( ColourAttr, (unsigned char) G);// >> 2);
	outp ( ColourAttr, (unsigned char) B);// >> 2);
}

//************************************************************
//                                                           *
//   Name    :	ProcessSample                                 *
//   Function: This attempts to "smooth" a sample out by     *
//             removing duplicate values and doing a linedraw*
//             to the next "level" ie.                       *
//             01 01 01 01 10 becomes.. 01 04 08 0c 10       *
//                                                           *
//   In   : pointer to buffer. last byte in buffer (nextbuf) *
//   Out  : buffer is smoothed                               *
//                                                           *
//************************************************************

void ProcessSample(int8 *pBuffer, float fnext)
{
	int32	i,i2,i3;
	float	step,v;

	pBuffer[312]=(uint8) fnext;
	pBuffer[313]=pBuffer[312]+1;

	i2=0; i=1;
	while (i<BUFFER_SIZE) {
		while (pBuffer[i2]==pBuffer[i]){
			i++;
		}
		if (i==BUFFER_SIZE+1) return;

		if (i2!=(i-1)){
			i3=i-i2;
			if (pBuffer[i2]<pBuffer[i]){
				step=(pBuffer[i]-pBuffer[i2])/(float) i3;	// get step inc
			} else {
				step=-((pBuffer[i2]-pBuffer[i])/(float) i3);
			}
			v=(float) pBuffer[i2];
			//i3--;
			if (i3>0){
			   for (;i3>0;i3--){
				   pBuffer[i2]=(uint8) v;
				   v=v+step;
				   i2++;
			   }
			   i++;
			}
		}else {
			i2++;
			i++;
		}

	}

}

/*************************************************************************
*
* FUNCTION: DSPOut()
*
* DESCRIPTION: Writes the value passed to this function to the DSP chip.
*
*************************************************************************/
void DSPOut(int IOBasePort, int WriteValue)
{
  // Wait until DSP is ready before writing the command
  while ((inp(IOBasePort + DSP_WRITE_PORT) & 0x80) != 0);

  outp(IOBasePort + DSP_WRITE_PORT, WriteValue);
  return;
}



//************************************************************
//                                                           *
//   Name    : SB_DMA_Alloc											 *
//   Function: Allocate real "dos" memory. Not protected mode*
//             memory, as we need a real 20bit address.		 *
//                                                           *
//   In   : *DMA = a DMA structure									 *
//          size = amount to malloc                          *
//   Out  : 0 = okay                                         *
//          1 = error                                        *
//                                                           *
//************************************************************

int SB_DMA_Alloc ( tSB_DMA_Data *dma, int size )
{
	uint32 addr;
	static union REGS reg;

	reg.x.eax = 0x0100; 			 			// DPMI Mem. Allocation
	reg.x.ebx = ((size*2)+15>>4);			// # of paragraphs to alloc.
	int386 ( 0x31, &reg,&reg );			// do it!!

	if ( reg.x.cflag ) return 1;			// allocate error
	dma->SB_DMA_Selector=reg.x.edx;  	// get the base selector
	addr=(reg.x.eax&0xFFFF)<<4; 	 		// get the real segment adress

	// to guarantee page continous, allocate twice the size of mem pool.
	// used in mikxmas player........
	if ( (addr>>16) != ((addr+size-1)>>16) )
		addr+=size;

	dma->SB_DMA_Page=addr>>16;		 		// Get the page
	dma->SB_DMA_Cont = (void *) addr;	// get the 20 bit real adress
	dma->SB_DMA_Length=size;		 		// store the size of the data
	return 0;
}



//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

int SB_DMA_Free ( tSB_DMA_Data *dma )
{
	static union REGS reg;
	reg.x.eax = 0x0101; 						// DPMI mem. dealloc.
	reg.x.edx = dma->SB_DMA_Selector;	// get the selector
	int386 ( 0x31, &reg, &reg );			// do it!!
	if ( reg.x.cflag ) return -1;       // error dealloc --> exit
	return 0;                           // if not....KEWL !!!
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

int32 ResetDSP ( void )
{
	uint8	al;

	if (SB_ON==0) return 0;


	outp	(BaseAddr+6,1);		// Start DSP reset
	delay (10);						// wait 10ms
	outp	(BaseAddr+6,0);		// end DSP reset

	WAITREAD (BaseAddr+14);		// DSP DATA AVAILABLE flag

	al=inp (BaseAddr+10);		// DSP Read Data
	if (al==0xaa){					// is sound blaster THERE?!?!?! 
      if (inp(BaseAddr + DSP_READ_PORT) == DSP_READY)
      {
         outp(BaseAddr + DSP_WRITE_PORT, DSP_VERSION);
         while ((inp(BaseAddr + DSP_DATA_AVAIL) & 0x80) == 0);
         DSP_Ver = inp(BaseAddr + DSP_READ_PORT);
         inp(BaseAddr + DSP_READ_PORT);
         return 0;
      }
	}
	SB_ON=0;							// we don't HAVE a sound blaster!
	return	1;
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void StopSample ( void )
{
	if (SB_ON==0) return;

	WAITWRITE (BaseAddr+0xc);
	outp (BaseAddr+0xc,0xd0);				// AL = Halt DMA Operation, 8-bit
	WAITWRITE (BaseAddr+0xc);
	outp (BaseAddr+0xc,0xda);				// AL = terminate 8-bit DMA Operation
	WAITWRITE (BaseAddr+0xc);
	outp (BaseAddr+0xc,0xd0);				// AL = Halt DMA Operation, 8-bit
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
#define PIC0_COMMAND_REG        0x20
#define PIC1_COMMAND_REG        0xA0
#define END_OF_INTERRUPT      0x0020

unsigned short int gDMABufNowPlaying,
		   gDMABufSize,       // See "WARNING #1" above.
		   gHighSpeed,
		   gIRQMaskSave0,
		   gIRQMaskSave1;
void interrupt (*gOldISRFuncPtr)(void);

void __interrupt __far SB_IRQ ( void )
{
  unsigned char InterruptStatus;
  

  SBcount++;
  SB_ready++;							// flag SB as "ready"

	interSetColour(0,ReD,0,0);

}

/*************************************************************************
*
* FUNCTION: DMAOutputISR()
*
* DESCRIPTION:  Interrupt service routine.  Every time the DSP chip finishes
*               playing half of the DMA buffer in auto-init mode, an
*               interrupt is generated, which invokes this routine.
*
*************************************************************************/
void interrupt DMAOutputISR(void)
{
  static char SecondToLastBufferPlayed = FALSE;
  int IntStatus;
	uint32	a;

	SB_ready++;							// flag SB as "ready"

	inp(BaseAddr + (int) DSP_DATA_AVAIL);  // Acknowledge interrupt (8-bit)
	outp(PIC_MODE, (int) PIC_END_OF_INT); // End of interrupt


	a=BufferList[0];					// rotate DMA buffers! so buffer "0" is free
	BufferList[0]=BufferList[1];
	BufferList[1]=a;
	return;
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
#define PIC0_MASK_REG           0x21
#define PIC1_MASK_REG           0xA1
void SetupIRQ ( void )
{
  // Save current ISR for IRQNumber, then give IRQNumber new ISR
  IRQSave = _dos_getvect( (IRQ + 8) );
  _dos_setvect(IRQ + 8, DMAOutputISR);

  // Save current interrupt mask and set new interrupt mask
  MaskSave = inp((int) PIC_MASK);
  IRQMask = ((int) 1 << IRQ); // Shift a 1 left IRQ of bits
  outp(PIC_MASK, (MaskSave & ~IRQMask)); // Enable previous AND new interrupts
}


void RestoreIRQ ( void )
{
  // Restore ISR and original IRQ vector
  outp(PIC_MASK, MaskSave);
  _dos_setvect( IRQ + 8, (void*)IRQSave);
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void CopySample( uint8 *pBuffer )
{
	if (SB_ON==0) return;						// copy samples??

	memcpy ((uint8*) (((uint8*)buffer1.SB_DMA_Cont)+BufferList[1]), pBuffer, BUFFER_SIZE);
}

void MixSample( uint8 *pSample )
{
	tSB_DMA_Data *buffer;
	int32			 i;
	uint8			 *pBuffer;
	uint32		 a,b,c,d;
	uint8				*p1,*p2,*p3,*p4;
	
	if (SB_ON==0) return;						// copy samples??

	p1=pBuffer2;
	p2=pBuffer3;
	p3=pBuffer4;
	p4=pBuffer5;

	pBuffer=(uint8*) (((uint8*)buffer1.SB_DMA_Cont)+BufferList[1]);
	for (i=0;i<BUFFER_SIZE;i++){
		b=(uint32)*p1++;
		b+=(uint32)*p2++;
		b+=(uint32)*p3++;
		b+=(uint32)*p4++;
		a=b/4;       
		a+=*pBuffer;
		if (a>255) a=255;
		*pBuffer++=((uint8)a);
	}
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void InitDMADSP ( void )
{
   outp(DMA8_MASK_REG, (int) (DMAchannel | 4));     // Disable DMA while prog.
   outp(DMA8_FF_REG,   (int) 0);                     // Clear the flip-flop

   outp(DMA8_MODE_REG, (int) (DMAchannel  | 0x58));  // 8-bit auto-init
	outp(SB_DMA_Length[DMAchannel], (int) (((BUFFER_SIZE*2) - 1) & 0xFF)); // LO byte of count
	outp(SB_DMA_Length[DMAchannel], (int) (((BUFFER_SIZE*2) - 1) >> 8));   // HI byte of count

	// Write the 20 bit address
	outp (SB_DMA_Page[DMAchannel]  ,(uint8) (((uint32) buffer1.SB_DMA_Cont>>16)&0xff) );	
	outp (SB_DMA_Offset[DMAchannel],(uint8) ((uint32) buffer1.SB_DMA_Cont&0xff));	
	outp (SB_DMA_Offset[DMAchannel],(uint8) ((uint32) buffer1.SB_DMA_Cont>>8)&0xff);	

	outp(DMA8_MASK_REG, DMAchannel);

  if(DSP_Ver < 4)
  {
	WAITWRITE (BaseAddr+0xc);
	outp (BaseAddr+0xc, 0x40);				//	Set TIME constant
	WAITWRITE (BaseAddr+0xc);
	outp (BaseAddr+0xc, ((uint8)TIMECONST));	
  }

  DSPOut(BaseAddr, 0x00D1);  // Must turn speaker ON before doing D/A conv.
}
//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void AllocateBuffers ( void )
{
	int32	i;
	uint8	*p1,*p2,*p3;

	SB_DMA_Alloc ( &buffer1, BUFFER_SIZE*2 );	// allocate a double buffer (624*2)

	p1=buffer1.SB_DMA_Cont;
	for (i=0;i<BUFFER_SIZE*2;i++){
		*p1++=0;
	}   
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void SetMixer(void)
{
  outp(BaseAddr + MIXER_ADDR, (int) MIC_VOLUME);
  outp(BaseAddr + MIXER_DATA, (int) 0x00);

  outp(BaseAddr + MIXER_ADDR, (int) VOICE_VOLUME);
  outp(BaseAddr + MIXER_DATA, (int) 0xFF);

  outp(BaseAddr + MIXER_ADDR, (int) MASTER_VOLUME);
  outp(BaseAddr + MIXER_DATA, (int) 0xFF);

  return;
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void Play( void )
{
  int Command;
	uint8	Mode=MONO;
	int	DMAMode=AUTO_INIT;


  if(DSP_Ver < 4) // SBPro (DSP ver 3.xx)
	{
		if (DMAMode == AUTO_INIT)
    	{
        DSPOut(BaseAddr, DSP_BLOCK_SIZE);
        DSPOut(BaseAddr, (int) ((BUFFER_SIZE - 1) & 0x00FF));
        DSPOut(BaseAddr, (int) ((BUFFER_SIZE - 1) >> 8));
        DSPOut(BaseAddr, 0x001C);  // AUTO INIT 8bit PCM
      }
      else
      {
        DSPOut(BaseAddr, 0x0014);  // SINGLE CYCLE 8bit PCM
        DSPOut(BaseAddr, (BUFFER_SIZE - 1) & 0x00FF);  // LO byte size
        DSPOut(BaseAddr, (BUFFER_SIZE - 1) >> 8);       // HI byte size
      }
	} else if(DSP_Ver == 4)// SB16 (DSP ver 4.xx)
  	{
    DSPOut(BaseAddr, 0x0041);  // DSP output transfer rate
    DSPOut(BaseAddr, (31200 & 0x0000FF00) >> 8 ); // Hi byte
    DSPOut(BaseAddr, (31200 & 0x000000FF) );        // Lo byte

	if (DMAMode == AUTO_INIT)
		DSPOut(BaseAddr, (8 == 8) ? 0x00C6 : 0x00B6);  // AUTO INIT 8/16 bit
	else
		DSPOut(BaseAddr, (8 == 8) ? 0x00C0 : 0x00B0);  // SINGLE CYCLE 8/16 bit

	if (8 == 8)
		DSPOut(BaseAddr, (Mode == MONO) ? 0x0000 : 0x0020); // 8bit MONO/STEREO
	else
		DSPOut(BaseAddr, (Mode == MONO) ? 0x0010 : 0x0030); // 16bit MONO/STEREO

	// Program number of samples to play
	DSPOut(BaseAddr, (int) ((BUFFER_SIZE - 1) & 0x00FF)); // LO byte
	DSPOut(BaseAddr, (int) ((BUFFER_SIZE - 1) >> 8));     // HI byte
  }

	BufferList[0]=0;								// playing buffer "0"
	BufferList[1]=(uint32) BUFFER_SIZE;		// next buffer
  return;
}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

int32 InitSB ( void )
{
	int32	i;
	uint8	*p1,*p2,*p3;

	pBuffer1=(uint8*)MyMalloc (1024);		// must ALWAYS malloc these!
	pBuffer2=(uint8*)MyMalloc (1024);
	pBuffer3=(uint8*)MyMalloc (1024);
	pBuffer4=(uint8*)MyMalloc (1024);
	pBuffer5=(uint8*)MyMalloc (1024);
	pBuffer6=(uint8*)MyMalloc (1024);
	pBuffer7=(uint8*)MyMalloc (1024);
	pBuffer8=(uint8*)MyMalloc (1024);

	for (i=0;i<1024;i++){
		pBuffer1[i]=0;
		pBuffer2[i]=0;
		pBuffer3[i]=0;
		pBuffer4[i]=0;
		pBuffer5[i]=0;
		pBuffer6[i]=0;
		pBuffer7[i]=0;
		pBuffer8[i]=0;
	}
	
	InitSID ();

	if (SB_ON==0) return 0;

	AllocateBuffers();
	TIMECONST=224;         		// = 31200 Hz (256-1000000/11000)
	BUFFER_SIZE=624;

	DMAchannel=3;           		// SoundBlaster DMA channel (default)
	BaseAddr=0x220;					// SoundBlaster BASE address (default)
	IRQ=7;

	DetectSB();							// detect using the BLASTER variable
	if (ResetDSP()==1) return 1;	// reset and test it its actually there!
	SetUpIRQ();							// set up IRQ
	InitDMADSP();						// Int DMA and DSP to play a BIG looping buffer
	SetMixer();							// set up mixer
	Play();

	return 0;
}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

int32	DisableSB ( void )
{
	MyFree(pBuffer1);
	MyFree(pBuffer2);
	MyFree(pBuffer3);
	MyFree(pBuffer4);
	MyFree(pBuffer5);
	MyFree(pBuffer6);
	MyFree(pBuffer7);
	MyFree(pBuffer8);

	FreeSID();						// free SID chip wav's etc.

	if (SB_ON==0) return 0; 	// Sound blaster already off?

	SB_DMA_Free ( &buffer1 );

   inp(BaseAddr + 0x000E);  	// Acknowledge the interrupt.
  	DSPOut(BaseAddr, DSP_HALT_SINGLE_CYCLE_DMA);  // Done playing, halt DMA
	ResetDSP ();
	RestoreIRQ();
	return 0;
}

//************************************************************
//                                                           *
//   Name    : DetectSB                                      *
//   Function: Detect the sound blaster via the "BLASTER"    *
//             environment variable                          *
//                                                           *
//   In   : none                                             *
//   Out  : IRQ, DMAChannel and BasePort are set.            *
//                                                           *
//************************************************************

void DetectSB( void )
{
	char	*pEnv;
	int32	i;
	int32	DMA,BPort,IIRQ;

	pEnv=(char*)getenv ("BLASTER");

	//
	// First search for the PORT using "A220" 
	//
	i=0;
	while (pEnv[i]!=0x0){
    	if (pEnv[i]=='A') break;
		i++;
	}
	if(pEnv[i]==0x0) return;
	i++;

	// now read it out...
	BPort=0;	
	while (pEnv[i]!=0x20){
		BPort=BPort*16;
		BPort=BPort+((uint8)pEnv[i]-48);
		i++;
	}
	BaseAddr=BPort;

	//
	// Now search for the DMA channel using "D3" 
	//
	i=0;
	while (pEnv[i]!=0x0){
    	if (pEnv[i]=='D') break;
		i++;
	}
	if(pEnv[i]==0x0) return;
	i++;

	// now read it out...
	DMA=0;	
	while (pEnv[i]!=0x20){
		DMA=DMA*10;
		DMA=DMA+ ((uint8)pEnv[i]-48);
		i++;
	}
	DMAchannel=DMA;


	//
	// Now search for the IRQ using "I7"
	//
	i=0;
	while (pEnv[i]!=0x0){
    	if (pEnv[i]=='I') break;
		i++;
	}
	if(pEnv[i]==0x0) return;
	i++;

	// now read it out...
	IRQ=0;	
	while (pEnv[i]!=0x20){
		IRQ=IRQ*10;
		IRQ=IRQ+ ((uint8)pEnv[i]-48);
		i++;
	}
}


//******************************************************************************
//
// SID sound emulation starts here!!
//
// Sample rate = 31200, sample sizes vary for each mode.
//
// To get new sample size we divide the RecordedSampleFreq/RequestedSampleFreq
// and then Xply by the original sample size. Then we simply scale the sample
// into its new size! Simple!
//
// MiddleC
// -------*SampleSize
//  Note
//
// 
// Value	Attack Rate		Decay/Release Rates
// ----- -----------    -------------------
//  0			2 ms					6 ms
//  1			8 ms             24 ms
//  2      16 ms             48 ms
//  3      24 ms             72 ms
//  4      38 ms            114 ms
//  5      56 ms            168 ms
//  6      68 ms            204 ms
//  7      80 ms            240 ms
//  8     100 ms            300 ms
//  9     240 ms            750 ms
// 10     500 ms            1.5 seconds
// 11     800 ms            2.4 seconds
// 12       1 second          3 seconds
// 13       3 seconds         9 seconds
// 14       5 seconds        15 seconds
// 15       8 seconds        25 seconds
//
//
// Frame rate     = 50Hz 
// ms			      = 1sec/1000
// ms in a frame	= 1sec/50	= 20ms
//
// So 20 Steps per frame. (This right?)
// We have 624 "clicks" per frame (sample rate is 31200)=31.2 clicks per ms
// 
// NOTE
// ----
// Since 25 seconds is 25000*31.2=780,000 steps (15/780000)(!) we
// can't use 16.16, but must use 8.24
//
//******************************************************************************

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

#define	MIDDLEC	4389.0f
//
// Actual HARDWARE
//
typedef	struct	tSIDREG{
		uint16	pitch;
		uint16	PulseSize;
		uint8		control;
		uint8		AD;
		uint8		SR;
}tSIDREG;



//
// OUR sid data.
//
typedef	struct	tSID{
		uint32	active;		// channel active?
		int32	   step;
		int32	   stepdelta;
		int32	   index;
		int32	   samplesize;
		uint8		*pSample;
		uint8		*pBuffer;
		uint32	ADSR;			   // ADSR "phase"
		int32		ADSRstep;	   // ADSR volume step(delta)
		int32		ADSRvolume;	   // Actual ADSR volume
		int32		TargetVolume;	// ADSR target volume
		tSIDREG	SidData;			// backup copy
}tSID;

		tSID		Sid[3];
		tSIDREG	*pSIDREG;
		uint32	TriSize;		// Triangle sample size
		uint8		*pTri;		// Triangle sample base
		uint32	SawSize;		// SAWtooth sample size
		uint8		*pSaw;		// SAWtooth sample base
		uint32	PulSize;		// Pulse sample size
		uint8		*pPul[3];	// pulse sample base 1 for each channel
		uint32	NoiSize;		// Pulse sample size
		uint8		*pNoi;		// noise sample base

// 15 volume levels (should really be 255)
#define	VOLUME	   15.0f
#define	ATTACK	   1
#define	DECAY		   2
#define	SUSTAIN	   3
#define	RELEASE	   4
#define	VOICE_OFF	0

	float		AttackRate[16]		={2,8,16,24,38,56,68,80,100,240,500,800,1000,3000,5000,8000};
	float		DecayReleaseRate[16]={6,24,48,72,114,168,204,240,300,750,1500,2400,3000,9000,15000,25000};
	int32		AttackStep[16];
	int32		DecayReleaseStep[16];

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void	InitADSR( void )
{
	float	step,a,b,base;
	int32	i;

	base=( ((float)BUFFER_SIZE) / (1000.0f/50) );
	for (i=0;i<16;i++){
		a=base*AttackRate[i];			      // get "clicks" count
		b=(VOLUME/a)*((float)0x1000000);			      // make 8.24
		AttackStep[i]=(int32)b;			      // and store...then cross our fingres!
	}

	for (i=0;i<16;i++){
		a=base*DecayReleaseRate[i];			// get "clicks" count
		b=(VOLUME/a)*((float)0x1000000);					// make 8.24
		DecayReleaseStep[i]=(int32)b;			
	}
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void InitSID ( void )
{
	TriSize=GetSize("midctri.raw");
	SawSize=GetSize("midcsaw.raw");
	PulSize=134;
	NoiSize=GetSize("noise.raw");

	pTri=(uint8*) WindLoad("midctri.raw");
	pSaw=(uint8*) WindLoad("midcsaw.raw");
	//pPul=(uint8*) WindLoad("midcpul.raw");
	pPul[0]=(uint8*) MyMalloc(134);								// single cycle at "middle C - C4"
	pPul[1]=(uint8*) MyMalloc(134);								// single cycle at "middle C - C4"
	pPul[2]=(uint8*) MyMalloc(134);								// single cycle at "middle C - C4"
	pNoi=(uint8*) WindLoad("noise.raw");

	Sid[0].pBuffer=pBuffer3;
	Sid[1].pBuffer=pBuffer4;
	Sid[2].pBuffer=pBuffer5;

	Sid[0].ADSR=VOICE_OFF;
	Sid[1].ADSR=VOICE_OFF;
	Sid[2].ADSR=VOICE_OFF;

	Sid[0].SidData.control=0;
	Sid[1].SidData.control=0;
	Sid[2].SidData.control=0;

	InitADSR();					// Set up adsr step rates
}
//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void FreeSID( void )
{
 	MyFree(pTri);
	MyFree(pSaw);
	MyFree(pPul[0]);
	MyFree(pPul[1]);
	MyFree(pPul[2]);
	MyFree(pNoi);
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************
void MakePulseWave(uint8 *pPul, int16 PulseSize)
{
	float		percent=(PulseSize&0xfff)/40.95f;		// get percentage left/right
	uint32	i,index,count;
	uint8		level;
	
	count=((PulSize/100.0f)*percent);

	level=0;
	index=0;
	for (i=0;i<count;i++){
		*pPul++=level;
		index++;
	}
	level=255/2;
	while (index<PulSize){
		*pPul++=level;
		index++;
	}
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void ProcessSIDChannel(tSID *pSid, int32	channel, int32 Regoffset)
{
	uint32	i;
	float		scale;
	uint32	size;


	pSIDREG=(tSIDREG*) &(Memory[0xfd40+Regoffset]);		// get base address of hardware

	// channel OFF?
	if ( ((pSIDREG->control&1)==0) AND (pSid->ADSR==VOICE_OFF)){
		pSid->active=0;                            
		return;
	}

	if ((pSIDREG->pitch!=pSid->SidData.pitch) OR				// has the user CHANGED the pitch?
      (pSIDREG->control!=pSid->SidData.control)){			// or the control byte

      
		pSid->active=1;

		if ((pSIDREG->control&1)!=(pSid->SidData.control&1)){		// channel JUST been started or stopped?
			if ((pSIDREG->control&1)==0){
				pSid->ADSR=RELEASE;											// Set RELEASE phase
				pSid->ADSRstep=DecayReleaseStep[(pSIDREG->SR&0xf)];	// get delta step!
				pSid->TargetVolume=0;
			}else{
				pSid->ADSR=ATTACK;												// Set ATTACK phase
				pSid->ADSRstep=AttackStep[(pSIDREG->AD>>4)];			// get delta step!
				pSid->ADSRvolume=0;										// set current volume!
				pSid->TargetVolume=15;
			}
		} 
		
      pSid->SidData.pitch=pSIDREG->pitch;
      pSid->SidData.control=pSIDREG->control;

      // get sample and sample size
      switch (pSIDREG->control&0xf0){
         case	16		:	pSid->pSample=pTri; 
								size = TriSize; 
								pSid->step=0;
								break;

         case	32		:	pSid->pSample=pSaw; 
								size = SawSize; 
								pSid->step=0;
								break;

         case	64		:	pSid->pSample=pPul[channel]; 
								size = PulSize; 
								pSid->step=0;
								break;

         case	128	:	pSid->pSample=pNoi; 
								size = NoiSize; 
								pSid->step=0;
								break;

			default		:	pSid->active=0;
      };
		pSid->samplesize = size<<16;

      scale= ( MIDDLEC/ ((float)pSIDREG->pitch) ) * size;	// get size sample SHOULD be
      pSid->stepdelta = (int32) ((size/scale)*65536.0f);		// get step delta

		pSid->SidData.AD=pSIDREG->AD;
		pSid->SidData.SR=pSIDREG->SR;
   }

	if (pSid->SidData.PulseSize!=pSIDREG->PulseSize){
		MakePulseWave(pPul[channel],pSIDREG->PulseSize);
		pSid->SidData.PulseSize=pSIDREG->PulseSize;
	}

}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void InitChannel1 ( void )
{
	ProcessSIDChannel(&Sid[0],0,0);	// process SID channel 1
}

//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void InitChannel2 ( void )
{
	ProcessSIDChannel(&Sid[1],1,7);	// process SID channel 1
}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void InitChannel3 ( void )
{
	ProcessSIDChannel(&Sid[2],2,14);	// process SID channel 1
}


//************************************************************
//                                                           *
//   Name    :                                               *
//   Function:                                               *
//                                                           *
//   In   :                                                  *
//   Out  :                                                  *
//                                                           *
//************************************************************

void ProcessSID( int32 index )
{
	tSID	*pSid;
	int	i;
	float	a,b;

  	if ((Memory[0xfd58]&0xf)==0){
		Sid[0].pBuffer[index]=0;			// No volume... so skip it
		Sid[1].pBuffer[index]=0;
		Sid[2].pBuffer[index]=0;
	}else{
	   pSid=&Sid[0];
	   for (i=0;i<3;i++){
	      if (pSid->active==1){

				// Process ADSR
			   switch (pSid->ADSR){
				   case	ATTACK:	if ((pSid->ADSRvolume>>24)==pSid->TargetVolume){
										   pSid->ADSR=DECAY;
										   pSid->ADSRstep=DecayReleaseStep[pSIDREG->AD&0xf];
										   pSid->TargetVolume=(pSIDREG->SR>>4);
									   }else{
											pSid->ADSRvolume+=pSid->ADSRstep;
										}
									   break;
				   case	DECAY:	if ((pSid->ADSRvolume>>24)==pSid->TargetVolume){
										   pSid->ADSR=SUSTAIN;
										   pSid->ADSRstep=0;
									   }else{
											pSid->ADSRvolume-=pSid->ADSRstep;
										}
									   break;

				   case	RELEASE:	if ((pSid->ADSRvolume>>24)==pSid->TargetVolume){
											pSid->active=0;
										   pSid->ADSR=VOICE_OFF;
									   }else{
											pSid->ADSRvolume-=pSid->ADSRstep;
										}
									   break;
			   }


				a=((float) pSid->pSample[(pSid->step>>16)]);
				b=((float) (pSid->ADSRvolume>>24));
				a=(((a*b)/15.0f) *(Memory[0xfd58]&0xf)/15.0f);
				
		      pSid->pBuffer[index]=(int32)a;
			   pSid->step+=pSid->stepdelta;
		      if (pSid->step > pSid->samplesize){
			      pSid->step-=pSid->samplesize;
		      }
	      }else{
		      pSid->pBuffer[index]=0;
		   }
		   pSid++;
	   }
	}
}


